# -*- coding: utf-8 -*-
"""
This module contains utility functions for downloading files.
"""
import os

import requests
from loguru import logger


def download_file(url: str, output_path: str):
    """
    A wrapper around requests.get to download a file from a URL.

    Args:
        url (str): The URL to download the file from.
        output_path (str): The path to save the downloaded file to.
    """

    logger.info(f"Downloading file from {url} to {output_path}")
    response = requests.get(url)
    response.raise_for_status()  # Check if the request was successful
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    with open(output_path, "wb") as f:
        f.write(response.content)
    logger.info(f"File downloaded to {output_path}")


def download_checkpoint(
    checkpoint_name: str, save_folder: str = "~/.local/mattersim/pretrained_models/"
):
    """
    Download a checkpoint from the Microsoft Mattersim repository.

    Args:
        checkpoint_name (str): The name of the checkpoint to download.
        save_folder (str): The local folder to save the checkpoint to.
    """

    GITHUB_CHECKPOINT_PREFIX = (
        "https://raw.githubusercontent.com/microsoft/mattersim/main/pretrained_models/"
    )
    checkpoint_url = GITHUB_CHECKPOINT_PREFIX + checkpoint_name.strip("/")
    save_path = os.path.join(
        os.path.expanduser(save_folder), checkpoint_name.strip("/")
    )
    download_file(checkpoint_url, save_path)



# -*- coding: utf-8 -*-
"""
Potential
"""
import os
import pickle
import random
import time
import warnings
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.distributed
import torch.nn as nn
from ase import Atoms
from ase.calculators.calculator import Calculator
from ase.constraints import full_3x3_to_voigt_6_stress
from ase.units import GPa
from deprecated import deprecated
from torch.optim import Adam
from torch.optim.lr_scheduler import ReduceLROnPlateau, StepLR
from torch_ema import ExponentialMovingAverage
from torch_geometric.loader import DataLoader
from torchmetrics import MeanMetric

from mattersim.datasets.utils.build import build_dataloader
from mattersim.forcefield.m3gnet.m3gnet import M3Gnet
from mattersim.jit_compile_tools.jit import compile_mode
from mattersim.utils.download_utils import download_checkpoint
from mattersim.utils.logger_utils import get_logger

rank = int(os.getenv("RANK", 0))
logger = get_logger()


@compile_mode("script")
class Potential(nn.Module):
    """
    A wrapper class for the force field model
    """

    def __init__(
        self,
        model,
        optimizer=None,
        scheduler: str = "StepLR",
        ema=None,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        allow_tf32=False,
        **kwargs,
    ):
        """
        Args:
            potential : a force field model
            lr : learning rate
            scheduler : a torch scheduler
            normalizer : an energy normalization module
        """
        super().__init__()
        self.model = model
        if optimizer is None:
            self.optimizer = Adam(
                self.model.parameters(), lr=kwargs.get("lr", 1e-3), eps=1e-7
            )
        else:
            self.optimizer = optimizer
        if not isinstance(scheduler, str):
            self.scheduler = scheduler
        elif scheduler == "StepLR":
            step_size = kwargs.get("step_size", 10)
            gamma = kwargs.get("gamma", 0.95)
            self.scheduler = StepLR(
                self.optimizer, step_size=step_size, gamma=gamma  # noqa: E501
            )
        elif scheduler == "ReduceLROnPlateau":
            factor = kwargs.get("factor", 0.8)
            patience = kwargs.get("patience", 50)
            self.scheduler = ReduceLROnPlateau(
                self.optimizer,
                mode="min",
                factor=factor,
                patience=patience,
                verbose=False,
            )
        else:
            raise NotImplementedError
        torch.backends.cuda.matmul.allow_tf32 = allow_tf32
        self.device = device
        self.to(device)

        if ema is None:
            self.ema = ExponentialMovingAverage(
                self.model.parameters(), decay=kwargs.get("ema_decay", 0.99)
            )
        else:
            self.ema = ema
        self.model_name = kwargs.get("model_name", "m3gnet")
        self.validation_metrics = kwargs.get(
            "validation_metrics", {"loss": 10000000.0}  # noqa: E501
        )
        self.last_epoch = kwargs.get("last_epoch", -1)
        self.description = kwargs.get("description", "")
        self.saved_name = ["loss", "MAE_energy", "MAE_force", "MAE_stress"]
        self.best_metric = 10000
        self.best_metric_epoch = 0
        self.rank = None

        self.use_finetune_label_loss = kwargs.get("use_finetune_label_loss", False)

    def freeze_reset_model(
        self,
        finetune_layers: int = -1,
        reset_head_for_finetune: bool = False,
    ):
        """
        Freeze the model in the fine-tuning process
        """
        if finetune_layers == -1:
            logger.info("fine-tuning all layers")
        elif finetune_layers >= 0 and finetune_layers < len(
            self.model.node_head.unified_encoder_layers
        ):
            logger.info(f"fine-tuning the last {finetune_layers} layers")
            for name, param in self.model.named_parameters():
                param.requires_grad = False

            # for energy head
            if finetune_layers > 0:
                for name, param in self.model.node_head.unified_encoder_layers[
                    -finetune_layers:
                ].named_parameters():
                    param.requires_grad = True
                for (
                    name,
                    param,
                ) in self.model.node_head.unified_final_invariant_ln.named_parameters():
                    param.requires_grad = True
                for (
                    name,
                    param,
                ) in self.model.node_head.unified_output_layer.named_parameters():
                    param.requires_grad = True
                for name, param in self.model.layer_norm.named_parameters():
                    param.requires_grad = True
            for name, param in self.model.lm_head_transform_weight.named_parameters():
                param.requires_grad = True
            for name, param in self.model.energy_out.named_parameters():
                param.requires_grad = True
            if reset_head_for_finetune:
                self.model.lm_head_transform_weight.reset_parameters()
                self.model.energy_out.reset_parameters()
        else:
            raise ValueError(
                "finetune_layers should be -1 or a positive integer,and less than the number of layers"  # noqa: E501
            )

    def finetune_mode(
        self,
        finetune_layers: int = -1,
        finetune_head: nn.Module = None,
        reset_head_for_finetune: bool = False,
        finetune_task_mean: float = 0.0,
        finetune_task_std: float = 1.0,
        use_finetune_label_loss: bool = False,
    ):
        """
        Set the model to fine-tuning mode
        finetune_layers: the layer to finetune, former layers will be frozen
                        if -1, all layers will be finetuned
        finetune_head: the head to finetune
        reset_head_for_finetune: whether to reset the original head
        """
        if self.model_name not in ["graphormer", "geomformer"]:
            logger.warning("Only graphormer and geomformer support freezing layers")
            return
        self.model.finetune_mode = True
        if finetune_head is None:
            logger.info("No finetune head is provided, using the original energy head")
        self.model.finetune_head = finetune_head
        self.model.finetune_task_mean = finetune_task_mean
        self.model.finetune_task_std = finetune_task_std
        self.freeze_reset_model(finetune_layers, reset_head_for_finetune)
        self.use_finetune_label_loss = use_finetune_label_loss

    def train_model(
        self,
        dataloader: Optional[list],
        val_dataloader,
        loss: torch.nn.modules.loss = torch.nn.MSELoss(),
        include_energy: bool = True,
        include_forces: bool = False,
        include_stresses: bool = False,
        force_loss_ratio: float = 1.0,
        stress_loss_ratio: float = 0.1,
        epochs: int = 100,
        early_stop_patience: int = 100,
        metric_name: str = "val_loss",
        wandb=None,
        save_checkpoint: bool = False,
        save_path: str = "./results/",
        ckpt_interval: int = 10,
        is_distributed: bool = False,
        need_to_load_data: bool = False,
        **kwargs,
    ):
        """
        Train model
        Args:
            dataloader: training data loader
            val_dataloader: validation data loader
            loss (torch.nn.modules.loss): loss object
            include_energy (bool) : whether to use energy as
                                    optimization targets
            include_forces (bool) : whether to use forces as
                                    optimization targets
            include_stresses (bool) : whether to use stresses as
                                      optimization targets
            force_loss_ratio (float): the ratio of forces in loss
            stress_loss_ratio (float): the ratio of stress in loss
            ckpt_interval (int): the interval to save checkpoints
            early_stop_patience (int): the patience for early stopping
            metric_name (str): the metric used for saving `best` checkpoints
                               and early stopping supported metrics:
                               `val_loss`, `val_mae_e`,
                               `val_mae_f`, `val_mae_s`
            sampler: used in distributed training
            is_distributed: whether to use DistributedDataParallel
            need_to_load_data: whether to load data from disk

        """
        self.idx = ["val_loss", "val_mae_e", "val_mae_f", "val_mae_s"].index(
            metric_name
        )
        if is_distributed:
            self.rank = torch.distributed.get_rank()
        logger.info(
            f"Number of trainable parameters: {sum(p.numel() for p in self.model.parameters() if p.requires_grad):,}"  # noqa: E501
        )
        for epoch in range(self.last_epoch + 1, epochs):
            logger.info(f"Epoch: {epoch} / {epochs}")
            if need_to_load_data:
                assert isinstance(dataloader, list)
                random.Random(kwargs.get("seed", 42) + epoch).shuffle(  # noqa: E501
                    dataloader
                )
                for idx, data_path in enumerate(dataloader):
                    with open(data_path, "rb") as f:
                        start = time.time()
                        train_data = pickle.load(f)
                    logger.info(
                        f"TRAIN: loading {data_path.split('/')[-2]}"
                        f"/{data_path.split('/')[-1]} dataset with "
                        f"{len(train_data)} data points, "
                        f"{len(train_data)} data points in total, "
                        f"time: {time.time() - start}"  # noqa: E501
                    )
                    # Distributed Sampling
                    atoms_train_sampler = (
                        torch.utils.data.distributed.DistributedSampler(
                            train_data,
                            seed=kwargs.get("seed", 42)
                            + idx * 131
                            + epoch,  # noqa: E501
                        )
                    )
                    train_dataloader = DataLoader(
                        train_data,
                        batch_size=kwargs.get("batch_size", 32),
                        shuffle=(atoms_train_sampler is None),
                        num_workers=0,
                        sampler=atoms_train_sampler,
                    )
                    metric = self.train_one_epoch(
                        train_dataloader,
                        epoch,
                        loss,
                        include_energy,
                        include_forces,
                        include_stresses,
                        force_loss_ratio,
                        stress_loss_ratio,
                        wandb,
                        is_distributed,
                        mode="train",
                        **kwargs,
                    )
                    del train_dataloader
                    del train_data
                    torch.cuda.empty_cache()
            else:
                metric = self.train_one_epoch(
                    dataloader,
                    epoch,
                    loss,
                    include_energy,
                    include_forces,
                    include_stresses,
                    force_loss_ratio,
                    stress_loss_ratio,
                    wandb,
                    is_distributed,
                    mode="train",
                    **kwargs,
                )
            if val_dataloader is not None:
                metric = self.train_one_epoch(
                    val_dataloader,
                    epoch,
                    loss,
                    include_energy,
                    include_forces,
                    include_stresses,
                    force_loss_ratio,
                    stress_loss_ratio,
                    wandb,
                    is_distributed,
                    mode="val",
                    **kwargs,
                )

            if isinstance(self.scheduler, ReduceLROnPlateau):
                self.scheduler.step(metric)
            else:
                self.scheduler.step()

            self.last_epoch = epoch

            self.validation_metrics = {
                "loss": metric[0],
                "MAE_energy": metric[1],
                "MAE_force": metric[2],
                "MAE_stress": metric[3],
            }
            if is_distributed:
                if self.save_model_ddp(
                    epoch,
                    early_stop_patience,
                    save_path,
                    metric_name,
                    save_checkpoint,
                    metric,
                    ckpt_interval,
                ):
                    break
            else:
                if self.save_model(
                    epoch,
                    early_stop_patience,
                    save_path,
                    metric_name,
                    save_checkpoint,
                    metric,
                    ckpt_interval,
                ):
                    break

    def save_model(
        self,
        epoch,
        early_stop_patience,
        save_path,
        metric_name,
        save_checkpoint,
        metric,
        ckpt_interval,
    ):
        with self.ema.average_parameters():
            try:
                best_model = torch.load(
                    os.path.join(save_path, "best_model.pth")  # noqa: E501
                )
                assert metric_name in [
                    "val_loss",
                    "val_mae_e",
                    "val_mae_f",
                    "val_mae_s",
                ], (
                    f"`{metric_name}` metric name not supported. "
                    "supported metrics: `val_loss`, `val_mae_e`, "
                    "`val_mae_f`, `val_mae_s`"
                )

                if (
                    save_checkpoint is True
                    and metric[self.idx]
                    < best_model["validation_metrics"][
                        self.saved_name[self.idx]
                    ]  # noqa: E501
                ):
                    self.save(os.path.join(save_path, "best_model.pth"))
                if epoch > best_model["last_epoch"] + early_stop_patience:
                    logger.info("Early stopping")
                    return True
                del best_model
            except BaseException:
                if save_checkpoint is True:
                    self.save(os.path.join(save_path, "best_model.pth"))

            if save_checkpoint is True and epoch % ckpt_interval == 0:
                self.save(os.path.join(save_path, f"ckpt_{epoch}.pth"))
            if save_checkpoint is True:
                self.save(os.path.join(save_path, "last_model.pth"))
            return False

    def save_model_ddp(
        self,
        epoch,
        early_stop_patience,
        save_path,
        metric_name,
        save_checkpoint,
        metric,
        ckpt_interval,
    ):
        with self.ema.average_parameters():
            assert metric_name in [
                "val_loss",
                "val_mae_e",
                "val_mae_f",
                "val_mae_s",
            ], (  # noqa: E501
                f"`{metric_name}` metric name not supported. "
                "supported metrics: `val_loss`, `val_mae_e`, "
                "`val_mae_f`, `val_mae_s`"
            )
            # Loading on multiple GPUs is too time consuming,
            # so this operation should not be performed.
            # Only save the model on GPU 0,
            # the model on each GPU should be exactly the same.
            if epoch > self.best_metric_epoch + early_stop_patience:
                logger.info("Early stopping")
                return True

            if metric[self.idx] < self.best_metric:
                self.best_metric = metric[self.idx]
                self.best_metric_epoch = epoch
                if save_checkpoint and self.rank == 0:
                    self.save(os.path.join(save_path, "best_model.pth"))
            if self.rank == 0 and save_checkpoint:
                if epoch % ckpt_interval == 0:
                    self.save(os.path.join(save_path, f"ckpt_{epoch}.pth"))
                self.save(os.path.join(save_path, "last_model.pth"))
            # torch.distributed.barrier()
            return False

    def test_model(
        self,
        val_dataloader,
        loss: torch.nn.modules.loss = torch.nn.MSELoss(),
        include_energy: bool = True,
        include_forces: bool = False,
        include_stresses: bool = False,
        wandb=None,
        **kwargs,
    ):
        """
        Test model performance on a given dataset
        """
        return self.train_one_epoch(
            val_dataloader,
            1,
            loss,
            include_energy,
            include_forces,
            include_stresses,
            1.0,
            0.1,
            wandb=wandb,
            mode="val",
            **kwargs,
        )

    def predict_properties(
        self,
        dataloader,
        include_forces: bool = False,
        include_stresses: bool = False,
        **kwargs,
    ) -> Tuple[List[float], List[np.ndarray], List[np.ndarray]]:
        """
        Predict properties (e.g., energies, forces) given a well-trained model
        Return: results tuple
            - results[0] (list[float]): a list of energies
            - results[1] (list[np.ndarray]): a list of atomic forces
            - results[2] (list[np.ndarray]): a list of stresses (in GPa)
        """
        self.model.eval()
        energies = []
        forces = []
        stresses = []
        for batch_idx, graph_batch in enumerate(dataloader):
            if self.model_name == "graphormer" or self.model_name == "geomformer":
                raise NotImplementedError
            else:
                graph_batch.to(self.device)
                input = batch_to_dict(graph_batch)
            result = self.forward(
                input,
                include_forces=include_forces,
                include_stresses=include_stresses,  # noqa: E501
            )
            if self.model_name == "graphormer" or self.model_name == "geomformer":
                raise NotImplementedError
            else:
                energies.extend(result["total_energy"].cpu().tolist())
                if include_forces:
                    forces_tuple = torch.split(
                        result["forces"].cpu().detach(),
                        graph_batch.num_atoms.cpu().tolist(),
                        dim=0,
                    )
                    for atomic_force in forces_tuple:
                        forces.append(np.array(atomic_force))
                if include_stresses:
                    stresses.extend(list(result["stresses"].cpu().detach().numpy()))

        return (energies, forces, stresses)

    # ============================

    def train_one_epoch(
        self,
        dataloader,
        epoch,
        loss,
        include_energy,
        include_forces,
        include_stresses,
        loss_f,
        loss_s,
        wandb,
        is_distributed=False,
        mode="train",
        log=True,
        **kwargs,
    ):
        start_time = time.time()
        loss_avg = MeanMetric().to(self.device)
        train_e_mae = MeanMetric().to(self.device)
        train_f_mae = MeanMetric().to(self.device)
        train_s_mae = MeanMetric().to(self.device)

        # scaler = torch.cuda.amp.GradScaler()

        if mode == "train":
            self.model.train()
        elif mode == "val":
            self.model.eval()

        for batch_idx, graph_batch in enumerate(dataloader):
            if self.model_name == "graphormer" or self.model_name == "geomformer":
                raise NotImplementedError
            else:
                graph_batch.to(self.device)
                input = batch_to_dict(graph_batch)
            if mode == "train":
                result = self.forward(
                    input,
                    include_forces=include_forces,
                    include_stresses=include_stresses,
                )
            elif mode == "val":
                with self.ema.average_parameters():
                    result = self.forward(
                        input,
                        include_forces=include_forces,
                        include_stresses=include_stresses,
                    )

            loss_, e_mae, f_mae, s_mae = self.loss_calc(
                graph_batch,
                result,
                loss,
                include_energy,
                include_forces,
                include_stresses,
                loss_f,
                loss_s,
            )

            # loss backward
            if mode == "train":
                self.optimizer.zero_grad()
                loss_.backward()
                nn.utils.clip_grad_norm_(
                    self.model.parameters(), 1.0, norm_type=2  # noqa: E501
                )
                self.optimizer.step()
                # scaler.scale(loss_).backward()
                # scaler.step(self.optimizer)
                # scaler.update()
                self.ema.update()

            loss_avg.update(loss_.detach())
            if include_energy:
                train_e_mae.update(e_mae.detach())
            if include_forces:
                train_f_mae.update(f_mae.detach())
            if include_stresses:
                train_s_mae.update(s_mae.detach())

        loss_avg_ = loss_avg.compute().item()
        if include_energy:
            e_mae = train_e_mae.compute().item()
        else:
            e_mae = 0
        if include_forces:
            f_mae = train_f_mae.compute().item()
        else:
            f_mae = 0
        if include_stresses:
            s_mae = train_s_mae.compute().item()
        else:
            s_mae = 0

        if log:
            logger.info(
                "%s: Loss: %.4f, MAE(e): %.4f, MAE(f): %.4f, MAE(s): %.4f, Time: %.2fs, lr: %.8f\n"  # noqa: E501
                % (
                    mode,
                    loss_avg.compute().item(),
                    e_mae,
                    f_mae,
                    s_mae,
                    time.time() - start_time,
                    self.scheduler.get_last_lr()[0],
                ),
            )

        if wandb and ((not is_distributed) or self.rank == 0):
            wandb.log(
                {
                    f"{mode}/loss": loss_avg_,
                    f"{mode}/mae_e": e_mae,
                    f"{mode}/mae_f": f_mae,
                    f"{mode}/mae_s": s_mae,
                    f"{mode}/lr": self.scheduler.get_last_lr()[0],
                    f"{mode}/mae_tot": e_mae + f_mae + s_mae,
                },
                step=epoch,
            )

        return (loss_avg_, e_mae, f_mae, s_mae)

    def loss_calc(
        self,
        graph_batch,
        result,
        loss,
        include_energy,
        include_forces,
        include_stresses,
        loss_f=1.0,
        loss_s=0.1,
    ):
        e_mae = 0.0
        f_mae = 0.0
        s_mae = 0.0
        loss_ = torch.tensor(0.0, device=self.device, requires_grad=True)

        if self.model_name == "graphormer" or self.model_name == "geomformer":
            raise NotImplementedError
        else:
            if include_energy:
                e_gt = graph_batch.energy / graph_batch.num_atoms
                e_pred = result["total_energy"] / graph_batch.num_atoms
                loss_ = loss_ + loss(e_pred, e_gt)
                e_mae = torch.nn.L1Loss()(e_pred, e_gt)
            if include_forces:
                f_gt = graph_batch.forces
                f_pred = result["forces"]
                loss_ = loss_ + loss(f_pred, f_gt) * loss_f
                f_mae = torch.nn.L1Loss()(f_pred, f_gt)
                # f_mae = torch.mean(torch.abs(f_pred - f_gt)).item()
            if include_stresses:
                s_gt = graph_batch.stress
                s_pred = result["stresses"]
                loss_ = loss_ + loss(s_pred, s_gt) * loss_s
                s_mae = torch.nn.L1Loss()(s_pred, s_gt)
                # s_mae = torch.mean(torch.abs((s_pred - s_gt))).item()
        return loss_, e_mae, f_mae, s_mae

    def get_properties(
        self,
        graph_batch,
        include_forces: bool = True,
        include_stresses: bool = True,
        **kwargs,
    ):
        """
        get energy, force and stress from a list of graph
        Args:
            graph_batch:
            include_forces (bool): whether to include force
            include_stresses (bool): whether to include stress
        Returns:
            results: a tuple, which consists of energies, forces and stress
        """
        warnings.warn(
            "This interface (get_properties) has been deprecated. "
            "Please use Potential.forward(input, include_forces, "
            "include_stresses) instead.",
            DeprecationWarning,
        )
        if self.model_name == "graphormer" or self.model_name == "geomformer":
            raise NotImplementedError
        else:
            graph_batch.to(self.device)
            input = batch_to_dict(graph_batch)
        result = self.forward(
            input,
            include_forces=include_forces,
            include_stresses=include_stresses,
            **kwargs,
        )
        # Warning: tuple
        if not include_forces and not include_stresses:
            return (result["total_energy"],)
        elif include_forces and not include_stresses:
            return (result["total_energy"], result["forces"])
        elif include_forces and include_stresses:
            return (result["total_energy"], result["forces"], result["stresses"])

    def forward(
        self,
        input: Dict[str, torch.Tensor],
        include_forces: bool = True,
        include_stresses: bool = True,
        dataset_idx: int = -1,
    ) -> Dict[str, torch.Tensor]:
        """
        get energy, force and stress from a list of graph
        Args:
            input: a dictionary contains all necessary info.
                   The `batch_to_dict` method could convert a graph_batch from
                   pyg dataloader to the input dictionary.
            include_forces (bool): whether to include force
            include_stresses (bool): whether to include stress
            dataset_idx (int): used for multi-head model, set to -1 by default
        Returns:
            results: a dictionary, which consists of energies,
                     forces and stresses
        """
        output = {}
        if self.model_name == "graphormer" or self.model_name == "geomformer":
            raise NotImplementedError
        else:
            strain = torch.zeros_like(input["cell"], device=self.device)
            volume = torch.linalg.det(input["cell"])
            if include_forces is True:
                input["atom_pos"].requires_grad_(True)
            if include_stresses is True:
                strain.requires_grad_(True)
                input["cell"] = torch.matmul(
                    input["cell"],
                    (torch.eye(3, device=self.device)[None, ...] + strain),
                )
                strain_augment = torch.repeat_interleave(
                    strain, input["num_atoms"], dim=0
                )
                input["atom_pos"] = torch.einsum(
                    "bi, bij -> bj",
                    input["atom_pos"],
                    (torch.eye(3, device=self.device)[None, ...] + strain_augment),
                )
                volume = torch.linalg.det(input["cell"])

            energies = self.model.forward(input, dataset_idx)
            output["total_energy"] = energies

            # Only take first derivative if only force is required
            if include_forces is True and include_stresses is False:
                grad_outputs: List[Optional[torch.Tensor]] = [
                    torch.ones_like(
                        energies,
                    )
                ]
                grad = torch.autograd.grad(
                    outputs=[
                        energies,
                    ],
                    inputs=[input["atom_pos"]],
                    grad_outputs=grad_outputs,
                    create_graph=self.model.training,
                )

                # Dump out gradient for forces
                force_grad = grad[0]
                if force_grad is not None:
                    forces = torch.neg(force_grad)
                    output["forces"] = forces

            # Take derivatives up to second order
            # if both forces and stresses are required
            if include_forces is True and include_stresses is True:
                grad_outputs: List[Optional[torch.Tensor]] = [
                    torch.ones_like(
                        energies,
                    )
                ]
                grad = torch.autograd.grad(
                    outputs=[
                        energies,
                    ],
                    inputs=[input["atom_pos"], strain],
                    grad_outputs=grad_outputs,
                    create_graph=self.model.training,
                )

                # Dump out gradient for forces and stresses
                force_grad = grad[0]
                stress_grad = grad[1]

                if force_grad is not None:
                    forces = torch.neg(force_grad)
                    output["forces"] = forces

                if stress_grad is not None:
                    stresses = (
                        1 / volume[:, None, None] * stress_grad / GPa
                    )  # 1/GPa = 160.21766208
                    output["stresses"] = stresses

        return output

    def save(self, save_path):
        dir_name = os.path.dirname(save_path)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        checkpoint = {
            "model_name": self.model_name,
            "model": self.model.module.state_dict()
            if hasattr(self.model, "module")
            else self.model.state_dict(),
            "model_args": self.model.module.get_model_args()
            if hasattr(self.model, "module")
            else self.model.get_model_args(),
            "optimizer": self.optimizer.state_dict(),
            "ema": self.ema.state_dict(),
            "scheduler": self.scheduler.state_dict(),
            "last_epoch": self.last_epoch,
            "validation_metrics": self.validation_metrics,
            "description": self.description,
        }
        torch.save(checkpoint, save_path)

    @classmethod
    def from_checkpoint(
        cls,
        load_path: str = None,
        *,
        model_name: str = "m3gnet",
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        load_training_state: bool = True,
        **kwargs,
    ):
        if model_name.lower() != "m3gnet":
            raise NotImplementedError

        checkpoint_folder = os.path.expanduser("~/.local/mattersim/pretrained_models")
        os.makedirs(checkpoint_folder, exist_ok=True)
        if (
            load_path is None
            or load_path.lower() == "mattersim-v1.0.0-1m.pth"
            or load_path.lower() == "mattersim-v1.0.0-1m"
        ):
            load_path = os.path.join(checkpoint_folder, "mattersim-v1.0.0-1M.pth")
            if not os.path.exists(load_path):
                logger.info(
                    "The pre-trained model is not found locally, "
                    "attempting to download it from the server."
                )
                download_checkpoint(
                    "mattersim-v1.0.0-1M.pth", save_folder=checkpoint_folder
                )
            logger.info(f"Loading the pre-trained {os.path.basename(load_path)} model")
        elif (
            load_path.lower() == "mattersim-v1.0.0-5m.pth"
            or load_path.lower() == "mattersim-v1.0.0-5m"
        ):
            load_path = os.path.join(checkpoint_folder, "mattersim-v1.0.0-5M.pth")
            if not os.path.exists(load_path):
                logger.info(
                    "The pre-trained model is not found locally, "
                    "attempting to download it from the server."
                )
                download_checkpoint(
                    "mattersim-v1.0.0-5M.pth", save_folder=checkpoint_folder
                )
            logger.info(f"Loading the pre-trained {os.path.basename(load_path)} model")
        else:
            logger.info("Loading the model from %s" % load_path)
        assert os.path.exists(load_path), f"Model file {load_path} not found"

        checkpoint = torch.load(load_path, map_location=device)

        assert checkpoint["model_name"] == model_name
        checkpoint["model_args"].update(kwargs)
        model = M3Gnet(device=device, **checkpoint["model_args"]).to(device)
        model.load_state_dict(checkpoint["model"], strict=False)

        if load_training_state:
            optimizer = Adam(model.parameters())
            scheduler = StepLR(optimizer, step_size=10, gamma=0.95)
            try:
                optimizer.load_state_dict(checkpoint["optimizer"])
            except BaseException:
                try:
                    optimizer.load_state_dict(checkpoint["optimizer"].state_dict())
                except BaseException:
                    optimizer = None
            try:
                scheduler.load_state_dict(checkpoint["scheduler"])
            except BaseException:
                try:
                    scheduler.load_state_dict(checkpoint["scheduler"].state_dict())
                except BaseException:
                    scheduler = "StepLR"
            try:
                last_epoch = checkpoint["last_epoch"]
                validation_metrics = checkpoint["validation_metrics"]
                description = checkpoint["description"]
            except BaseException:
                last_epoch = -1
                validation_metrics = {"loss": 0.0}
                description = ""
            try:
                ema = ExponentialMovingAverage(model.parameters(), decay=0.99)
                ema.load_state_dict(checkpoint["ema"])
            except BaseException:
                ema = None
        else:
            optimizer = None
            scheduler = "StepLR"
            last_epoch = -1
            validation_metrics = {"loss": 0.0}
            description = ""
            ema = None

        model.eval()

        del checkpoint

        return cls(
            model,
            optimizer=optimizer,
            ema=ema,
            scheduler=scheduler,
            device=device,
            model_name=model_name,
            last_epoch=last_epoch,
            validation_metrics=validation_metrics,
            description=description,
            **kwargs,
        )

    @deprecated(version="1.0.0", reason="Please use from_checkpoint instead.")
    def load(
        load_path: str = None,
        *,
        model_name: str = "m3gnet",
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        args: Dict = None,
        load_training_state: bool = True,
        **kwargs,
    ):
        if model_name.lower() != "m3gnet":
            raise NotImplementedError

        checkpoint_folder = os.path.expanduser("~/.local/mattersim/pretrained_models")
        os.makedirs(checkpoint_folder, exist_ok=True)
        if (
            load_path is None
            or load_path.lower() == "mattersim-v1.0.0-1m.pth"
            or load_path.lower() == "mattersim-v1.0.0-1m"
        ):
            load_path = os.path.join(checkpoint_folder, "mattersim-v1.0.0-1M.pth")
            if not os.path.exists(load_path):
                logger.info(
                    "The pre-trained model is not found locally, "
                    "attempting to download it from the server."
                )
                download_checkpoint(
                    "mattersim-v1.0.0-1M.pth", save_folder=checkpoint_folder
                )
            logger.info(f"Loading the pre-trained {os.path.basename(load_path)} model")
        elif (
            load_path.lower() == "mattersim-v1.0.0-5m.pth"
            or load_path.lower() == "mattersim-v1.0.0-5m"
        ):
            load_path = os.path.join(checkpoint_folder, "mattersim-v1.0.0-5M.pth")
            if not os.path.exists(load_path):
                logger.info(
                    "The pre-trained model is not found locally, "
                    "attempting to download it from the server."
                )
                download_checkpoint(
                    "mattersim-v1.0.0-5M.pth", save_folder=checkpoint_folder
                )
            logger.info(f"Loading the pre-trained {os.path.basename(load_path)} model")
        else:
            logger.info("Loading the model from %s" % load_path)

        assert os.path.exists(load_path), f"Model file {load_path} not found"

        checkpoint = torch.load(load_path, map_location=device)

        assert checkpoint["model_name"] == model_name
        checkpoint["model_args"].update(kwargs)
        model = M3Gnet(device=device, **checkpoint["model_args"]).to(device)
        model.load_state_dict(checkpoint["model"], strict=False)

        if load_training_state:
            optimizer = Adam(model.parameters())
            scheduler = StepLR(optimizer, step_size=10, gamma=0.95)
            try:
                optimizer.load_state_dict(checkpoint["optimizer"])
            except BaseException:
                try:
                    optimizer.load_state_dict(checkpoint["optimizer"].state_dict())
                except BaseException:
                    optimizer = None
            try:
                scheduler.load_state_dict(checkpoint["scheduler"])
            except BaseException:
                try:
                    scheduler.load_state_dict(checkpoint["scheduler"].state_dict())
                except BaseException:
                    scheduler = "StepLR"
            try:
                last_epoch = checkpoint["last_epoch"]
                validation_metrics = checkpoint["validation_metrics"]
                description = checkpoint["description"]
            except BaseException:
                last_epoch = -1
                validation_metrics = {"loss": 0.0}
                description = ""
            try:
                ema = ExponentialMovingAverage(model.parameters(), decay=0.99)
                ema.load_state_dict(checkpoint["ema"])
            except BaseException:
                ema = None
        else:
            optimizer = None
            scheduler = "StepLR"
            last_epoch = -1
            validation_metrics = {"loss": 0.0}
            description = ""
            ema = None

        model.eval()

        del checkpoint

        return Potential(
            model,
            optimizer=optimizer,
            ema=ema,
            scheduler=scheduler,
            device=device,
            model_name=model_name,
            last_epoch=last_epoch,
            validation_metrics=validation_metrics,
            description=description,
            **kwargs,
        )

    def set_description(self, description):
        self.description = description

    def get_description(self):
        return self.description


def batch_to_dict(graph_batch, model_type="m3gnet", device="cuda"):
    if model_type == "m3gnet":
        # TODO: key_list
        atom_pos = graph_batch.atom_pos
        cell = graph_batch.cell
        pbc_offsets = graph_batch.pbc_offsets
        atom_attr = graph_batch.atom_attr
        edge_index = graph_batch.edge_index
        three_body_indices = graph_batch.three_body_indices
        num_three_body = graph_batch.num_three_body
        num_bonds = graph_batch.num_bonds
        num_triple_ij = graph_batch.num_triple_ij
        num_atoms = graph_batch.num_atoms
        num_graphs = graph_batch.num_graphs
        num_graphs = torch.tensor(num_graphs)
        batch = graph_batch.batch

        # Resemble input dictionary
        input = {}
        input["atom_pos"] = atom_pos
        input["cell"] = cell
        input["pbc_offsets"] = pbc_offsets
        input["atom_attr"] = atom_attr
        input["edge_index"] = edge_index
        input["three_body_indices"] = three_body_indices
        input["num_three_body"] = num_three_body
        input["num_bonds"] = num_bonds
        input["num_triple_ij"] = num_triple_ij
        input["num_atoms"] = num_atoms
        input["num_graphs"] = num_graphs
        input["batch"] = batch
    elif model_type == "graphormer" or model_type == "geomformer":
        raise NotImplementedError
    else:
        raise NotImplementedError

    return input


@deprecated(version="1.0.0", reason="Please use MatterSimCalculator instead.")
class DeepCalculator(Calculator):
    """
    Deep calculator based on ase Calculator
    """

    implemented_properties = ["energy", "free_energy", "forces", "stress"]

    def __init__(
        self,
        potential: Potential,
        args_dict: dict = {},
        compute_stress: bool = True,
        stress_weight: float = 1.0,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        **kwargs,
    ):
        """
        Args:
            potential (Potential): m3gnet.models.Potential
            compute_stress (bool): whether to calculate the stress
            stress_weight (float): the stress weight.
            **kwargs:
        """
        super().__init__(**kwargs)
        self.potential = potential
        self.compute_stress = compute_stress
        self.stress_weight = stress_weight
        self.args_dict = args_dict
        self.device = device

    @classmethod
    def from_checkpoint(cls, load_path: str, **kwargs):
        potential = Potential.from_checkpoint(load_path, **kwargs)
        return cls(potential, **kwargs)

    @classmethod
    def from_potential(cls, potential: Potential, **kwargs):
        return cls(potential, **kwargs)

    def calculate(
        self,
        atoms: Optional[Atoms] = None,
        properties: Optional[list] = None,
        system_changes: Optional[list] = None,
    ):
        """
        Args:
            atoms (ase.Atoms): ase Atoms object
            properties (list): list of properties to calculate
            system_changes (list): monitor which properties of atoms were
                changed for new calculation. If not, the previous calculation
                results will be loaded.
        Returns:
        """

        all_changes = [
            "positions",
            "numbers",
            "cell",
            "pbc",
            "initial_charges",
            "initial_magmoms",
        ]

        properties = properties or ["energy"]
        system_changes = system_changes or all_changes
        super().calculate(
            atoms=atoms, properties=properties, system_changes=system_changes
        )

        self.args_dict["batch_size"] = 1
        self.args_dict["only_inference"] = 1
        cutoff = (
            self.potential.model.model_args["cutoff"]
            if self.potential.model_name == "m3gnet"
            else 5.0
        )
        threebody_cutoff = (
            self.potential.model.model_args["threebody_cutoff"]
            if self.potential.model_name == "m3gnet"
            else 4.0
        )

        dataloader = build_dataloader(
            [atoms],
            model_type=self.potential.model_name,
            cutoff=cutoff,
            threebody_cutoff=threebody_cutoff,
            **self.args_dict,
        )
        for graph_batch in dataloader:
            # Resemble input dictionary
            if (
                self.potential.model_name == "graphormer"
                or self.potential.model_name == "geomformer"
            ):
                raise NotImplementedError
            else:
                graph_batch = graph_batch.to(self.device)
                input = batch_to_dict(graph_batch)

            result = self.potential.forward(
                input, include_forces=True, include_stresses=self.compute_stress
            )
            if (
                self.potential.model_name == "graphormer"
                or self.potential.model_name == "geomformer"
            ):
                raise NotImplementedError
            else:
                self.results.update(
                    energy=result["total_energy"].detach().cpu().numpy()[0],
                    free_energy=result["total_energy"].detach().cpu().numpy()[0],
                    forces=result["forces"].detach().cpu().numpy(),
                )
            if self.compute_stress:
                self.results.update(
                    stress=self.stress_weight
                    * full_3x3_to_voigt_6_stress(
                        result["stresses"].detach().cpu().numpy()[0]
                    )
                )


class MatterSimCalculator(Calculator):
    """
    Deep calculator based on ase Calculator
    """

    implemented_properties = ["energy", "free_energy", "forces", "stress"]

    def __init__(
        self,
        potential: Potential = None,
        args_dict: dict = {},
        compute_stress: bool = True,
        stress_weight: float = GPa,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        **kwargs,
    ):
        """
        Args:
            potential (Potential): m3gnet.models.Potential
            compute_stress (bool): whether to calculate the stress
            stress_weight (float): the stress weight.
            **kwargs:
        """
        super().__init__(**kwargs)
        if potential is None:
            self.potential = Potential.from_checkpoint(device=device, **kwargs)
        else:
            self.potential = potential
        self.compute_stress = compute_stress
        self.stress_weight = stress_weight
        self.args_dict = args_dict
        self.device = device

    @classmethod
    def from_checkpoint(cls, load_path: str, **kwargs):
        potential = Potential.from_checkpoint(load_path, **kwargs)
        return cls(potential=potential, **kwargs)

    @classmethod
    def from_potential(cls, potential: Potential, **kwargs):
        return cls(potential=potential, **kwargs)

    @deprecated(
        version="1.0.0", reason="Please use from_checkpoint or from_potential instead."
    )
    def load(
        load_path: str = None,
        *,
        model_name: str = "m3gnet",
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        args: Dict = None,
        load_training_state: bool = True,
        args_dict: dict = {},
        compute_stress: bool = True,
        stress_weight: float = GPa,
        **kwargs,
    ):
        potential = Potential.load(
            load_path=load_path,
            model_name=model_name,
            device=device,
            args=args,
            load_training_state=load_training_state,
        )
        return MatterSimCalculator(
            potential=potential,
            args_dict=args_dict,
            compute_stress=compute_stress,
            stress_weight=stress_weight,
            device=device,
            **kwargs,
        )

    def calculate(
        self,
        atoms: Optional[Atoms] = None,
        properties: Optional[list] = None,
        system_changes: Optional[list] = None,
    ):
        """
        Args:
            atoms (ase.Atoms): ase Atoms object
            properties (list): list of properties to calculate
            system_changes (list): monitor which properties of atoms were
                changed for new calculation. If not, the previous calculation
                results will be loaded.
        Returns:
        """

        all_changes = [
            "positions",
            "numbers",
            "cell",
            "pbc",
            "initial_charges",
            "initial_magmoms",
        ]

        properties = properties or ["energy"]
        system_changes = system_changes or all_changes
        super().calculate(
            atoms=atoms, properties=properties, system_changes=system_changes
        )

        self.args_dict["batch_size"] = 1
        self.args_dict["only_inference"] = 1
        cutoff = (
            self.potential.model.model_args["cutoff"]
            if self.potential.model_name == "m3gnet"
            else 5.0
        )
        threebody_cutoff = (
            self.potential.model.model_args["threebody_cutoff"]
            if self.potential.model_name == "m3gnet"
            else 4.0
        )

        dataloader = build_dataloader(
            [atoms],
            model_type=self.potential.model_name,
            cutoff=cutoff,
            threebody_cutoff=threebody_cutoff,
            **self.args_dict,
        )
        for graph_batch in dataloader:
            # Resemble input dictionary
            if (
                self.potential.model_name == "graphormer"
                or self.potential.model_name == "geomformer"
            ):
                raise NotImplementedError
            else:
                graph_batch = graph_batch.to(self.device)
                input = batch_to_dict(graph_batch)

            result = self.potential.forward(
                input, include_forces=True, include_stresses=self.compute_stress
            )
            if (
                self.potential.model_name == "graphormer"
                or self.potential.model_name == "geomformer"
            ):
                raise NotImplementedError
            else:
                self.results.update(
                    energy=result["total_energy"].detach().cpu().numpy()[0],
                    free_energy=result["total_energy"].detach().cpu().numpy()[0],
                    forces=result["forces"].detach().cpu().numpy(),
                )
            if self.compute_stress:
                self.results.update(
                    stress=self.stress_weight
                    * full_3x3_to_voigt_6_stress(
                        result["stresses"].detach().cpu().numpy()[0]
                    )
                )



# -*- coding: utf-8 -*-
from functools import lru_cache

import numpy as np
import torch
from ase import Atoms
from torch_geometric.data import Data


@torch.jit.script
def convert_to_single_emb(x, offset: int = 512):
    feature_num = x.size(1) if len(x.size()) > 1 else 1
    feature_offset = 1 + torch.arange(0, feature_num * offset, offset, dtype=torch.long)
    x = x + feature_offset
    return x


class AtomCalDataset:
    def __init__(
        self,
        atom_list: list[Atoms],
        energies: list[float] = None,
        forces: list[np.ndarray] = None,
        stresses: list[np.ndarray] = None,
        finetune_task_label: list = None,
    ):
        self.data = self._preprocess(
            atom_list,
            energies,
            forces,
            stresses,
            finetune_task_label,
        )

    def _preprocess(
        self,
        atom_list,
        energies: list[float] = None,
        forces: list[np.ndarray] = None,
        stresses: list[np.ndarray] = None,
        finetune_task_label: list = None,
        use_ase_energy: bool = False,
        use_ase_force: bool = False,
        use_ase_stress: bool = False,
    ):
        data_list = []
        for i, (atom, energy, force, stress) in enumerate(
            zip(atom_list, energies, forces, stresses)
        ):
            item_dict = atom.todict()
            item_dict["info"] = {}
            if energy is None:
                energy = 0
            if force is None:
                force = np.zeros([len(atom), 3])
            if stress is None:
                stress = np.zeros([3, 3])
            try:
                energy = atom.get_total_energy() if use_ase_energy else energy
                force = (
                    atom.get_forces(apply_constraint=False) if use_ase_force else force
                )
                stress = atom.get_stress(voigt=False) if use_ase_stress else stress
            except Exception as e:
                RuntimeError(f"Error in {i}th data: {e}")

            if finetune_task_label is not None:
                item_dict["finetune_task_label"] = finetune_task_label[i]
            else:
                item_dict["finetune_task_label"] = 0

            item_dict["info"]["energy"] = energy
            item_dict["info"]["stress"] = stress  # * 160.2176621
            item_dict["forces"] = force
            data_list.append(item_dict)

        return data_list

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        item = self.data[idx]
        return preprocess_atom_item(item, idx)

    def __len__(self):
        return len(self.data)


def preprocess_atom_item(item, idx):
    # numbers = item.pop("numbers")
    numbers = item["numbers"]
    item["x"] = torch.tensor(numbers, dtype=torch.long).unsqueeze(-1)
    # positions = item.pop("positions")
    positions = item["positions"]
    item["pos"] = torch.tensor(positions, dtype=torch.float64)
    item["cell"] = torch.tensor(item["cell"], dtype=torch.float64)
    item["pbc"] = torch.tensor(item["pbc"], dtype=torch.bool)
    item["idx"] = idx
    item["y"] = torch.tensor([item["finetune_task_label"]])
    item["total_energy"] = torch.tensor([item["info"]["energy"]], dtype=torch.float64)
    item["stress"] = torch.tensor(item["info"]["stress"], dtype=torch.float64)
    item["forces"] = torch.tensor(item["forces"], dtype=torch.float64)

    item = Data(**item)

    x = item.x

    item.x = convert_to_single_emb(x)

    return item



# -*- coding: utf-8 -*-
import os
import sys

from loguru import logger

handlers = {}


def get_logger():
    if not handlers:
        logger.remove()
        handlers["console"] = logger.add(
            sys.stdout,
            colorize=True,
            filter=log_filter,
            enqueue=True,
        )

    return logger


def log_filter(record):
    if record["level"].name != "INFO":
        return True

    if "RANK" not in os.environ or int(os.environ["RANK"]) == 0:
        return True
    else:
        return False



# -*- coding: utf-8 -*-
import unittest

import numpy as np
from ase import Atoms
from ase.calculators.emt import EMT
from phonopy import Phonopy

from mattersim.applications.phonon import PhononWorkflow


class PhononTestCase(unittest.TestCase):
    def setUp(self):
        # Create an example structure for testing
        a = 1.786854996
        positions = [
            (1.78685500, 1.78685500, 1.78685500),
            (2.68028249, 2.68028249, 2.68028249),
        ]
        cell = [(0, a, a), (a, 0, a), (a, a, 0)]

        self.atoms = Atoms("C2", positions=positions, cell=cell, pbc=True)

        # Create an conventional cell for testing
        a2 = a * 2
        positions2 = [
            (0, 0, 0),
            (0, a2 / 2, a2 / 2),
            (a2 / 2, 0, a2 / 2),
            (a2 / 2, a2 / 2, 0),
            (a2 / 4, a2 / 4, a2 / 4),
            (a2 / 4, 3 * a2 / 4, 3 * a2 / 4),
            (3 * a2 / 4, a2 / 4, 3 * a2 / 4),
            (3 * a2 / 4, 3 * a2 / 4, a2 / 4),
        ]

        cell2 = [(a2, 0, 0), (0, a2, 0), (0, 0, a2)]

        self.atoms_conv = Atoms("C8", positions=positions2, cell=cell2, pbc=True)

        self.calculator = EMT()
        self.atoms.calc = self.calculator
        self.atoms_conv.calc = self.calculator

    def test_phonon(self):
        phononworkflow = PhononWorkflow(self.atoms, work_dir="/tmp/diamond")
        has_imaginary, phonon = phononworkflow.run()

        self.assertTrue(has_imaginary)
        self.assertIsInstance(phonon, Phonopy)

    def test_phonon_supercell(self):
        supercell_matrix = np.array([[4, 0, 0], [0, 4, 0], [0, 0, 4]])
        qpoints_mesh = np.array([12, 12, 12])
        phononworkflow = PhononWorkflow(
            self.atoms,
            work_dir="/tmp/diamond",
            supercell_matrix=supercell_matrix,
            qpoints_mesh=qpoints_mesh,
        )
        has_imaginary, phonon = phononworkflow.run()

        self.assertTrue(has_imaginary)
        self.assertIsInstance(phonon, Phonopy)

    def test_phonon_prim(self):
        phononworkflow = PhononWorkflow(
            self.atoms_conv, work_dir="/tmp/diamond_conv", find_prim=True
        )
        has_imaginary, phonon = phononworkflow.run()
        has_imaginary, phonon = phononworkflow.run()

        self.assertTrue(has_imaginary)
        self.assertIsInstance(phonon, Phonopy)


if __name__ == "__main__":
    unittest.main()



import os
import uuid
from collections import defaultdict
from typing import List, Union

import pandas as pd
from ase import Atoms
from ase.filters import Filter
from ase.optimize.optimize import Optimizer
from ase.units import GPa
from loguru import logger
from pymatgen.io.ase import AseAtomsAdaptor
from tqdm import tqdm

from mattersim.applications.relax import Relaxer


def relax(
    atoms_list: List[Atoms],
    *,
    optimizer: Union[str, Optimizer] = "FIRE",
    filter: Union[str, Filter, None] = None,
    constrain_symmetry: bool = False,
    fix_axis: Union[bool, List[bool]] = False,
    pressure_in_GPa: float = None,
    fmax: float = 0.01,
    steps: int = 500,
    work_dir: str = str(uuid.uuid4()),
    save_csv: str = "results.csv.gz",
    **kwargs,
) -> dict:
    """
    Relax a list of atoms structures.

    Args:
        atoms_list (List[Atoms]): List of ASE Atoms objects.
        optimizer (Union[str, Optimizer]): The optimizer to use. Default is "FIRE".
        filter (Union[str, Filter, None]): The filter to use.
        constrain_symmetry (bool): Whether to constrain symmetry. Default is False.
        fix_axis (Union[bool, List[bool]]): Whether to fix the axis. Default is False.
        pressure_in_GPa (float): Pressure in GPa to use for relaxation.
        fmax (float): Maximum force tolerance for relaxation. Default is 0.01.
        steps (int): Maximum number of steps for relaxation. Default is 500.
        work_dir (str): Working directory for the calculations.
            Default is a UUID with timestamp.
        save_csv (str): Save the results to a CSV file. Default is `results.csv.gz`.

    Returns:
        pd.DataFrame: DataFrame containing the relaxed results.
    """
    params_filter = {}

    if pressure_in_GPa:
        params_filter["scalar_pressure"] = (
            pressure_in_GPa * GPa
        )  # convert GPa to eV/Angstrom^3
        filter = "ExpCellFilter" if filter is None else filter
    elif filter:
        params_filter["scalar_pressure"] = 0.0

    relaxer = Relaxer(
        optimizer=optimizer,
        filter=filter,
        constrain_symmetry=constrain_symmetry,
        fix_axis=fix_axis,
    )

    relaxed_results = defaultdict(list)
    for atoms in tqdm(atoms_list, total=len(atoms_list), desc="Relaxing structures"):
        converged, relaxed_atoms = relaxer.relax(
            atoms,
            params_filter=params_filter,
            fmax=fmax,
            steps=steps,
        )
        relaxed_results["converged"].append(converged)
        relaxed_results["structure"].append(
            AseAtomsAdaptor.get_structure(relaxed_atoms).to_json()
        )
        relaxed_results["energy"].append(relaxed_atoms.get_potential_energy())
        relaxed_results["energy_per_atom"].append(
            relaxed_atoms.get_potential_energy() / len(relaxed_atoms)
        )
        relaxed_results["forces"].append(relaxed_atoms.get_forces())
        relaxed_results["stress"].append(relaxed_atoms.get_stress(voigt=False))
        relaxed_results["stress_GPa"].append(
            relaxed_atoms.get_stress(voigt=False) / GPa
        )

        logger.info(f"Relaxed structure: {relaxed_atoms}")

    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    logger.info(f"Saving the results to {os.path.join(work_dir, save_csv)}")
    df = pd.DataFrame(relaxed_results)
    df.to_csv(os.path.join(work_dir, save_csv), index=False, mode="a")
    return relaxed_results



# -*- coding: utf-8 -*-
import logging
import pathlib
import sys
from typing import Dict, Tuple, Union

if sys.version_info[1] >= 8:
    from typing import Final
else:
    from typing_extensions import Final

# This is a weird hack to avoid Intel MKL issues on the cluster
import ase.data

# when this is called as a subprocess of a process that has itself initialized
# PyTorch. Since numpy gets imported later anyway for dataset stuff,
# this shouldn't affect performance.
import numpy as np  # noqa: F401
import torch

from .jit import script

# Denote meta_data_keys
TWO_BODY_CUTOFF: Final[str] = "two_body_cutoff"
HAS_THREE_BODY: Final[str] = "has_three_body"
THREE_BODY_CUTOFF: Final[str] = "three_body_cutoff"
N_SPECIES_KEY: Final[str] = "n_species"
TYPE_NAMES_KEY: Final[str] = "type_names"
JIT_BAILOUT_KEY: Final[str] = "_jit_bailout_depth"
JIT_FUSION_STRATEGY: Final[str] = "_jit_fusion_strategy"
TF32_KEY: Final[str] = "allow_tf32"

_ALL_METADATA_KEYS = [
    TWO_BODY_CUTOFF,
    HAS_THREE_BODY,
    THREE_BODY_CUTOFF,
    N_SPECIES_KEY,
    TYPE_NAMES_KEY,
    JIT_BAILOUT_KEY,
    JIT_FUSION_STRATEGY,
    TF32_KEY,
]


def _compile_for_deploy(model):
    model.eval()

    if not isinstance(model, torch.jit.ScriptModule):
        print("Non TorchScript model detected,JIT  compiling the model ....")
        model = script(model)
    else:
        print(
            "Model provided is already a TorchScript model, "
            "return as it is."  # noqa: E501
        )
    return model


def load_deployed_model(
    model_path: Union[pathlib.Path, str],
    device: Union[str, torch.device] = "cpu",
    freeze: bool = True,
) -> Tuple[torch.jit.ScriptModule, Dict[str, str]]:
    r"""Load a deployed model.
    Args:
        model_path: the path to the deployed model's ``.pth`` file.
    Returns:
        model, metadata dictionary
    """
    metadata = {k: "" for k in _ALL_METADATA_KEYS}
    try:
        model = torch.jit.load(
            model_path, map_location=device, _extra_files=metadata
        )  # noqa: E501
    except RuntimeError as e:
        raise ValueError(
            f"{model_path} does not seem to be a deployed RL4CSP model file. "
            f"Did you forget to deploy it? \n\n(Underlying error: {e})"
        )

    # Confirm its TorchScript
    assert isinstance(model, torch.jit.ScriptModule)

    # Make sure we're in eval mode
    model.eval()
    # Freeze on load:
    if freeze and hasattr(model, "training"):
        # hasattr is how torch checks whether model is unfrozen
        # only freeze if already unfrozen
        model = torch.jit.freeze(model)

    # Everything we store right now is ASCII, so decode for printing
    metadata = {k: v.decode("ascii") for k, v in metadata.items()}

    # JIT strategy
    strategy = metadata.get(JIT_FUSION_STRATEGY, "")

    if strategy != "":
        strategy = [e.split(",") for e in strategy.split(";")]
        strategy = [(e[0], int(e[1])) for e in strategy]
    else:
        print(
            "Missing information: JIT strategy, "
            "loading deployed model fails !"  # noqa: E501
        )
        exit()

    # JIT bailout
    jit_bailout: int = metadata.get(JIT_BAILOUT_KEY, "")
    if jit_bailout == "":
        print(
            "Missing information: JIT_BAILOUT_KEY, "
            "loading deployed model fails !"  # noqa: E501
        )
        exit()

    # JIT allow_tf32
    jit_allow_tf32: int = metadata.get(TF32_KEY, "")
    if jit_allow_tf32 == "":
        print("Missing information: TF32_KEY, loading deployed model fails !")
        exit()

    return model, metadata


def deploy(
    model,
    is_m3gnet_pretrained=False,
    is_m3gnet_multi_head_pretrained=False,
    metadata=None,
    deployed_model_name="deployed.pth",
    device="cpu",
):
    # Compile model
    complied_model = _compile_for_deploy(model)

    # Use default metadata dictionary for pretrained models
    if is_m3gnet_pretrained:
        metadata = {}

        # Do set differences get atomic numbers
        full_atomic_numbers = set(np.arange(1, 95, 1))
        discard_atomic_numbers = set(np.arange(84, 89, 1))
        covered_atomic_numbers = list(
            full_atomic_numbers.difference(discard_atomic_numbers)
        )
        type_names = []
        for atomic_num in covered_atomic_numbers:
            type_names.append(ase.data.chemical_symbols[atomic_num])
        metadata[TWO_BODY_CUTOFF] = str(5.0)
        metadata[HAS_THREE_BODY] = str(True)
        metadata[THREE_BODY_CUTOFF] = str(4.0)
        metadata[N_SPECIES_KEY] = str(89)
        metadata[TYPE_NAMES_KEY] = " ".join(type_names)
        metadata[JIT_BAILOUT_KEY] = str(2)
        metadata[JIT_FUSION_STRATEGY] = ";".join(
            "%s,%i" % e for e in [("DYNAMIC", 3)]  # noqa: E501
        )
        metadata[TF32_KEY] = str(int(0))

    # TODO: Add default meta keys for m3gent_multi_head models
    # elif is_m3gnet_multi_head_pretrained:

    else:
        # Missing fields in meta data triggers failing compilation
        metadata_keys = metadata.keys
        for _ALL_METADATA_KEY in _ALL_METADATA_KEYS:
            if _ALL_METADATA_KEY not in metadata_keys:
                logging.info(
                    "Miss metadata key: "
                    + _ALL_METADATA_KEY
                    + " model deploying fails!"
                )
                exit()
        # Missing metadata values, other than JIT compile information,
        # triggers failing compilation
        for i in range(len(metadata_keys) - 3):
            if metadata[metadata_keys[i]].empty():
                logging.info(
                    "metadata with key "
                    + metadata_keys
                    + "not set, model deploying fails!"
                )
                exit()
        # Set default JIT compile information is values are not set
        if (
            metadata["JIT_BAILOUT_KEY"].empty()
            or metadata[JIT_FUSION_STRATEGY].empty()
            or metadata[TF32_KEY].empty()
        ):
            metadata[JIT_BAILOUT_KEY] = str(2)
            metadata[JIT_FUSION_STRATEGY] = ";".join(
                "%s,%i" % e for e in [("DYNAMIC", 3)]
            )
            metadata[TF32_KEY] = str(int(0))

    # Deploy model with full information
    # Confirm its TorchScript
    assert isinstance(complied_model, torch.jit.ScriptModule)
    if device != "cuda":
        complied_model = complied_model.cpu()

    torch.jit.save(complied_model, deployed_model_name, _extra_files=metadata)

    return complied_model, metadata



# -*- coding: utf-8 -*-
"""jit.py is used to compile model with jit and modified from
    https://github.com/e3nn/e3nn/blob/main/e3nn/util/jit.py
"""

import copy
import inspect
import warnings
from typing import Optional

import torch
from opt_einsum_fx import jitable
from torch import fx

_RL4CSP_COMPILE_MODE = "__rl4csp_compile_mode__"
_VALID_MODES = ("trace", "script", "unsupported", None)
_MAKE_TRACING_INPUTS = "_make_tracing_inputs"


def compile_mode(mode: str):
    """Decorator to set the compile mode of a module.

    Parameters
    ----------
        mode : str
            'script', 'trace', or None
    """
    if mode not in _VALID_MODES:
        raise ValueError("Invalid compile mode")

    def decorator(obj):
        if not (inspect.isclass(obj) and issubclass(obj, torch.nn.Module)):
            raise TypeError(
                "@rl4csp.mattersim.forcefield.jit.compile_mode can only "
                "decorate classes derived from torch.nn.Module"
            )
        setattr(obj, _RL4CSP_COMPILE_MODE, mode)
        return obj

    return decorator


def get_compile_mode(mod: torch.nn.Module) -> str:
    """Get the compilation mode of a module.

    Parameters
    ----------
        mod : torch.nn.Module

    Returns
    -------
    'script', 'trace', or None if the module was not decorated with
                       @compile_mode
    """
    if hasattr(mod, _RL4CSP_COMPILE_MODE):
        mode = getattr(mod, _RL4CSP_COMPILE_MODE)
    else:
        mode = getattr(type(mod), _RL4CSP_COMPILE_MODE, None)
    if mode is None and isinstance(mod, fx.GraphModule):
        mode = "script"
    assert mode in _VALID_MODES, "Invalid compile mode `%r`" % mode
    return mode


def compile(
    mod: torch.nn.Module,
    n_trace_checks: int = 1,
    script_options: dict = None,
    trace_options: dict = None,
    in_place: bool = True,
):
    """Recursively compile a module and all submodules according
       to their decorators.

    (Sub)modules without decorators will be unaffected.

    Parameters
    ----------
        mod : torch.nn.Module
            The module to compile. The module will have its submodules
            compiled replaced in-place.
        n_trace_checks : int, default = 1
            How many random example inputs to generate when tracing a module.
            Must be at least one in order to have a tracing input.
            Extra example inputs will be passed to ``torch.jit.trace``
            to confirm that the traced copmute graph doesn't change.
        script_options : dict, default = {}
            Extra kwargs for ``torch.jit.script``.
        trace_options : dict, default = {}
            Extra kwargs for ``torch.jit.trace``.

    Returns
    -------
    Returns the compiled module.
    :param trace_options:
    :param script_options:
    :param n_trace_checks:
    :param mod:
    :param in_place:
    """
    script_options = script_options or {}
    trace_options = trace_options or {}

    mode = get_compile_mode(mod)
    if mode == "unsupported":
        raise NotImplementedError(
            f"{type(mod).__name__} does not support TorchScript compilation"
        )

    if not in_place:
        mod = copy.deepcopy(mod)
    # TODO: debug logging
    assert n_trace_checks >= 1
    # == recurse to children ==
    # This allows us to trace compile submodules of modules we are going to
    # script
    for submod_name, submod in mod.named_children():
        setattr(
            mod,
            submod_name,
            compile(
                submod,
                n_trace_checks=n_trace_checks,
                script_options=script_options,
                trace_options=trace_options,
                in_place=True,
                # since we deepcopied the module above, we can do inplace
            ),
        )
    # == Compile this module now ==
    if mode == "script":
        if isinstance(mod, fx.GraphModule):
            mod = jitable(mod)
        mod = torch.jit.script(mod, **script_options)
    elif mode == "trace":
        # These are always modules, so we're always using trace_module
        # We need tracing inputs:
        check_inputs = get_tracing_inputs(
            mod,
            n_trace_checks,
        )
        assert len(check_inputs) >= 1, "Must have at least one tracing input."
        # Do the actual trace
        mod = torch.jit.trace_module(
            mod,
            inputs=check_inputs[0],
            check_inputs=check_inputs,
            **trace_options,  # noqa: E501
        )
    return mod


def get_tracing_inputs(
    mod: torch.nn.Module,
    n: int = 1,
    device: Optional[torch.device] = None,
    dtype: Optional[torch.dtype] = None,
):
    """Get random tracing inputs for ``mod``.

    First checks if ``mod`` has a ``_make_tracing_inputs`` method.
    If so, calls it with ``n`` as the single argument and returns its results.

    Otherwise, attempts to infer the input signature of the module using
    ``e3nn.util._argtools._get_io_irreps``.

    Parameters
    ----------
        mod : torch.nn.Module
        n : int, default = 1
            A hint for how many inputs are wanted. Usually n will be returned,
            but modules don't necessarily have to.
        device : torch.device
            The device to do tracing on. If `None` (default), will be guessed.
        dtype : torch.dtype
            The dtype to trace with. If `None` (default), will be guessed.

    Returns
    -------
    list of dict
        Tracing inputs in the format of ``torch.jit.trace_module``:
        dicts mapping method names like ``'forward'`` to tuples of arguments.
    """
    # Avoid circular imports
    from ._argtools import (
        _get_device,
        _get_floating_dtype,
        _get_io_irreps,
        _rand_args,
        _to_device_dtype,
    )

    # - Get inputs -
    if hasattr(mod, _MAKE_TRACING_INPUTS):
        # This returns a trace_module style dict of method names to test inputs
        trace_inputs = mod._make_tracing_inputs(n)
        assert isinstance(trace_inputs, list)
        for d in trace_inputs:
            assert isinstance(
                d, dict
            ), "_make_tracing_inputs must return a list of dict[str, tuple]"
            assert all(
                isinstance(k, str) and isinstance(v, tuple)
                for k, v in d.items()  # noqa: E501
            ), "_make_tracing_inputs must return a list of dict[str, tuple]"
    else:
        # Try to infer. This will throw if it can't.
        irreps_in, _ = _get_io_irreps(
            mod, irreps_out=[None]
        )  # we're only trying to infer inputs
        trace_inputs = [{"forward": _rand_args(irreps_in)} for _ in range(n)]
    # - Put them on the right device -
    if device is None:
        device = _get_device(mod)
    if dtype is None:
        dtype = _get_floating_dtype(mod)
    # Move them
    trace_inputs = _to_device_dtype(trace_inputs, device, dtype)
    return trace_inputs


def trace_module(
    mod: torch.nn.Module,
    inputs: dict = None,
    check_inputs: list = None,
    in_place: bool = True,
):
    """Trace a module.

    Identical signature to ``torch.jit.trace_module``, but first recursively
    compiles ``mod`` using ``compile``.

    Parameters
    ----------
        mod : torch.nn.Module
        inputs : dict
        check_inputs : list of dict
    Returns
    -------
    Traced module.
    """
    check_inputs = check_inputs or []

    # Set the compile mode for mod, temporarily
    old_mode = getattr(mod, _RL4CSP_COMPILE_MODE, None)
    if old_mode is not None and old_mode != "trace":
        warnings.warn(
            f"Trying to trace a module of type {type(mod).__name__} marked "
            "with @compile_mode != 'trace', expect errors!"
        )
    setattr(mod, _RL4CSP_COMPILE_MODE, "trace")

    # If inputs are provided, set make_tracing_input temporarily
    old_make_tracing_input = None
    if inputs is not None:
        old_make_tracing_input = getattr(mod, _MAKE_TRACING_INPUTS, None)
        setattr(
            mod,
            _MAKE_TRACING_INPUTS,
            lambda num: ([inputs] + check_inputs),  # noqa: E501
        )

    # Compile
    out = compile(mod, in_place=in_place)

    # Restore old values, if we had them
    if old_mode is not None:
        setattr(mod, _RL4CSP_COMPILE_MODE, old_mode)
    if old_make_tracing_input is not None:
        setattr(mod, _MAKE_TRACING_INPUTS, old_make_tracing_input)
    return out


def trace(
    mod: torch.nn.Module,
    example_inputs: tuple = None,
    check_inputs: list = None,
    in_place: bool = True,
):
    """Trace a module.

    Identical signature to ``torch.jit.trace``, but first recursively compiles
    ``mod`` using :func:``compile``.

    Parameters
    ----------
        mod : torch.nn.Module
        example_inputs : tuple
        check_inputs : list of tuple
    Returns
    -------
    Traced module.
    """
    check_inputs = check_inputs or []

    return trace_module(
        mod=mod,
        inputs=(
            {"forward": example_inputs}
            if example_inputs is not None
            else None  # noqa: E501
        ),
        check_inputs=[{"forward": c} for c in check_inputs],
        in_place=in_place,
    )


def script(mod, in_place: bool = True):
    """Script a module.

    Like ``torch.jit.script``, but first recursively compiles ``mod``
    using :func:``compile``.

    Parameters
    ----------
        mod : torch.nn.Module
    Returns
    -------
    Scripted module.
    """
    # Set the compile mode for mod, temporarily
    old_mode = getattr(mod, _RL4CSP_COMPILE_MODE, None)
    if old_mode is not None and old_mode != "script":
        warnings.warn(
            f"Trying to script a module of type {type(mod).__name__} marked "
            "with @compile_mode != 'script', expect errors! "
        )
    setattr(mod, _RL4CSP_COMPILE_MODE, "script")

    # Compile
    out = compile(mod, in_place=in_place)

    # Restore old values, if we had them
    if old_mode is not None:
        setattr(mod, _RL4CSP_COMPILE_MODE, old_mode)

    return out



import os
import uuid
from collections import defaultdict
from typing import List

import numpy as np
import pandas as pd
import yaml
from ase import Atoms
from loguru import logger
from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from tqdm import tqdm

from mattersim.applications.phonon import PhononWorkflow
from mattersim.cli.applications.relax import relax


def phonon(
    atoms_list: List[Atoms],
    *,
    find_prim: bool = False,
    work_dir: str = str(uuid.uuid4()),
    save_csv: str = "results.csv.gz",
    amplitude: float = 0.01,
    supercell_matrix: np.ndarray = None,
    qpoints_mesh: np.ndarray = None,
    max_atoms: int = None,
    enable_relax: bool = False,
    **kwargs,
) -> dict:
    """
    Predict phonon properties for a list of atoms.

    Args:
        atoms_list (List[Atoms]): List of ASE Atoms objects.
        find_prim (bool, optional): If find the primitive cell and use it
            to calculate phonon. Default to False.
        work_dir (str, optional): workplace path to contain phonon result.
            Defaults to data + chemical_symbols + 'phonon'
        amplitude (float, optional): Magnitude of the finite difference to
            displace in force constant calculation, in Angstrom. Defaults
            to 0.01 Angstrom.
        supercell_matrix (nd.array, optional): Supercell matrix for constr
            -uct supercell, priority over than max_atoms. Defaults to None.
        qpoints_mesh (nd.array, optional): Qpoint mesh for IBZ integral,
            priority over than max_atoms. Defaults to None.
        max_atoms (int, optional): Maximum atoms number limitation for the
            supercell generation. If not set, will automatic generate super
            -cell based on symmetry. Defaults to None.
        enable_relax (bool, optional): Whether to relax the structure before
            predicting phonon properties. Defaults to False.
    """
    phonon_results = defaultdict(list)

    for atoms in tqdm(
        atoms_list, total=len(atoms_list), desc="Predicting phonon properties"
    ):
        if enable_relax:
            relaxed_results = relax(
                [atoms],
                constrain_symmetry=True,
                work_dir=work_dir,
                save_csv=save_csv.replace(".csv", "_relax.csv"),
            )
            structure = Structure.from_str(relaxed_results["structure"][0], fmt="json")
            _atoms = AseAtomsAdaptor.get_atoms(structure)
            _atoms.calc = atoms.calc
            atoms = _atoms
        ph = PhononWorkflow(
            atoms=atoms,
            find_prim=find_prim,
            work_dir=work_dir,
            amplitude=amplitude,
            supercell_matrix=supercell_matrix,
            qpoints_mesh=qpoints_mesh,
            max_atoms=max_atoms,
        )
        has_imaginary, phonon = ph.run()
        phonon_results["has_imaginary"].append(has_imaginary)
        # phonon_results["phonon"].append(phonon)
        phonon_results["phonon_band_plot"].append(
            os.path.join(os.path.abspath(work_dir), f"{atoms.symbols}_phonon_band.png")
        )
        phonon_results["phonon_dos_plot"].append(
            os.path.join(os.path.abspath(work_dir), f"{atoms.symbols}_phonon_dos.png")
        )
        os.rename(
            os.path.join(os.path.abspath(work_dir), "band.yaml"),
            os.path.join(os.path.abspath(work_dir), f"{atoms.symbols}_band.yaml"),
        )
        os.rename(
            os.path.join(os.path.abspath(work_dir), "phonopy_params.yaml"),
            os.path.join(
                os.path.abspath(work_dir), f"{atoms.symbols}_phonopy_params.yaml"
            ),
        )
        os.rename(
            os.path.join(os.path.abspath(work_dir), "total_dos.dat"),
            os.path.join(os.path.abspath(work_dir), f"{atoms.symbols}_total_dos.dat"),
        )
        phonon_results["phonon_band"].append(
            yaml.safe_load(
                open(
                    os.path.join(
                        os.path.abspath(work_dir), f"{atoms.symbols}_band.yaml"
                    ),
                    "r",
                )
            )
        )
        phonon_results["phonopy_params"].append(
            yaml.safe_load(
                open(
                    os.path.join(
                        os.path.abspath(work_dir),
                        f"{atoms.symbols}_phonopy_params.yaml",
                    ),
                    "r",
                )
            )
        )
        phonon_results["total_dos"].append(
            np.loadtxt(
                os.path.join(
                    os.path.abspath(work_dir), f"{atoms.symbols}_total_dos.dat"
                ),
                comments="#",
            )
        )

    if not os.path.exists(work_dir):
        os.makedirs(work_dir)

    logger.info(f"Saving the results to {os.path.join(work_dir, save_csv)}")
    df = pd.DataFrame(phonon_results)
    df.to_csv(
        os.path.join(work_dir, save_csv.replace(".csv", "_phonon.csv")),
        index=False,
        mode="a",
    )
    return phonon_results



# -*- coding: utf-8 -*-
import sys
from typing import Dict, List, Union

from ase import Atoms, units
from ase.calculators.calculator import Calculator
from ase.filters import ExpCellFilter, FrechetCellFilter, Filter
from ase.optimize import BFGS, FIRE
from ase.optimize.optimize import Optimizer
from loguru import logger
from tqdm import tqdm

from mattersim.datasets.utils.build import build_dataloader
from mattersim.forcefield.potential import Potential


class DummyBatchCalculator(Calculator):
    def __init__(self):
        super().__init__()

    def calculate(self, atoms=None, properties=None, system_changes=None):
        pass

    def get_potential_energy(self, atoms=None):
        return atoms.info["total_energy"]

    def get_forces(self, atoms=None):
        return atoms.arrays["forces"]

    def get_stress(self, atoms=None):
        return units.GPa * atoms.info["stress"]


class BatchRelaxer(object):
    """BatchRelaxer is a class for batch structural relaxation.
    It is more efficient than Relaxer when relaxing a large number of structures."""

    SUPPORTED_OPTIMIZERS = {"BFGS": BFGS, "FIRE": FIRE}
    SUPPORTED_FILTERS = {
        "EXPCELLFILTER": ExpCellFilter,
        "FRECHETCELLFILTER": FrechetCellFilter,
    }

    def __init__(
        self,
        potential: Potential,
        optimizer: Union[str, type[Optimizer]] = "FIRE",
        filter: Union[type[Filter], str, None] = None,
        fmax: float = 0.05,
        max_natoms_per_batch: int = 512,
        max_n_steps: int = 1_000_000,
    ):
        self.potential = potential
        self.device = potential.device
        if isinstance(optimizer, str):
            if optimizer.upper() not in self.SUPPORTED_OPTIMIZERS:
                raise ValueError(f"Unsupported optimizer: {optimizer}")
            self.optimizer = self.SUPPORTED_OPTIMIZERS[optimizer.upper()]
        elif issubclass(optimizer, Optimizer):
            self.optimizer = optimizer
        else:
            raise ValueError(f"Unsupported optimizer: {optimizer}")
        if isinstance(filter, str):
            if filter.upper() not in self.SUPPORTED_FILTERS:
                raise ValueError(f"Unsupported filter: {filter}")
            self.filter = self.SUPPORTED_FILTERS[filter.upper()]
        elif filter is None or issubclass(filter, Filter):
            self.filter = filter
        else:
            raise ValueError(f"Unsupported filter: {filter}")
        self.fmax = fmax
        self.max_natoms_per_batch = max_natoms_per_batch
        self.optimizer_instances: List[Optimizer] = []
        self.is_active_instance: List[bool] = []
        self.finished = False
        self.total_converged = 0
        self.trajectories: Dict[int, List[Atoms]] = {}
        self.max_n_steps = max_n_steps 

    def insert(self, atoms: Atoms):
        atoms.calc = DummyBatchCalculator()
        optimizer_instance = self.optimizer(
            self.filter(atoms) if self.filter else atoms
        )
        optimizer_instance.fmax = self.fmax
        optimizer_instance.nsteps = 0
        self.optimizer_instances.append(optimizer_instance)
        self.is_active_instance.append(True)

    def step_batch(self):
        atoms_list = []
        for idx, opt in enumerate(self.optimizer_instances):
            if self.is_active_instance[idx]:
                atoms_list.append(opt.atoms)

        # Note: we use a batch size of len(atoms_list)
        # because we only want to run one batch at a time
        dataloader = build_dataloader(
            atoms_list, batch_size=len(atoms_list), only_inference=True
        )
        energy_batch, forces_batch, stress_batch = self.potential.predict_properties(
            dataloader, include_forces=True, include_stresses=True
        )

        counter = 0
        self.finished = True
        for idx, opt in enumerate(self.optimizer_instances):
            if self.is_active_instance[idx]:
                # Set the properties so the dummy calculator can
                # return them within the optimizer step
                opt.atoms.info["total_energy"] = energy_batch[counter]
                opt.atoms.arrays["forces"] = forces_batch[counter]
                opt.atoms.info["stress"] = stress_batch[counter]
                try:
                    self.trajectories[opt.atoms.info["structure_index"]].append(
                        opt.atoms.copy()
                    )
                except KeyError:
                    self.trajectories[opt.atoms.info["structure_index"]] = [
                        opt.atoms.copy()
                    ]

                opt.step()
                opt.nsteps += 1
                # Get gradient for convergence check
                # Note: gradient = -forces for the optimizable object
                gradient = opt.optimizable.get_gradient()
                if opt.converged(gradient) or opt.nsteps >= self.max_n_steps:
                    self.is_active_instance[idx] = False
                    self.total_converged += 1
                    if self.total_converged % 100 == 0:
                        logger.info(f"Relaxed {self.total_converged} structures.")
                else:
                    self.finished = False
                counter += 1

        # remove inactive instances
        self.optimizer_instances = [
            opt
            for opt, active in zip(self.optimizer_instances, self.is_active_instance)
            if active
        ]
        self.is_active_instance = [True] * len(self.optimizer_instances)

    def relax(
        self,
        atoms_list: List[Atoms],
    ) -> Dict[int, List[Atoms]]:
        self.trajectories = {}
        self.tqdmcounter = tqdm(total=len(atoms_list), file=sys.stdout)
        pointer = 0
        atoms_list_ = []
        for i in range(len(atoms_list)):
            atoms_list_.append(atoms_list[i].copy())
            atoms_list_[i].info["structure_index"] = i

        while (
            pointer < len(atoms_list) or not self.finished
        ):  # While there are unfinished instances or atoms left to insert
            while pointer < len(atoms_list) and (
                sum([len(opt.atoms) for opt in self.optimizer_instances])
                + len(atoms_list[pointer])
                <= self.max_natoms_per_batch
            ):
                # While there are enough n_atoms slots in the
                # batch and we have not reached the end of the list.
                self.insert(
                    atoms_list_[pointer]
                )  # Insert new structure to fire instances
                self.tqdmcounter.update(1)
                pointer += 1
            self.step_batch()
        self.tqdmcounter.close()

        return self.trajectories



# -*- coding: utf-8 -*-
from distutils.core import Extension, setup

import numpy
from Cython.Build import cythonize

package = Extension(
    "threebody_indices",
    ["threebody_indices.pyx"],
    include_dirs=[numpy.get_include()],  # noqa: E501
)
setup(ext_modules=cythonize([package]))

# usage:
# python setup.py build_ext --inplace



# -*- coding: utf-8 -*-
"""
Nequip
"""

import logging
from typing import Optional

import numpy as np
import torch
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, Hyperparameter, Kernel


def solver(
    X, y, regressor: Optional[str] = "NormalizedGaussianProcess", **kwargs
):  # noqa: E501
    if regressor == "GaussianProcess":
        return gp(X, y, **kwargs)
    elif regressor == "NormalizedGaussianProcess":
        return normalized_gp(X, y, **kwargs)
    else:
        raise NotImplementedError(f"{regressor} is not implemented")


def normalized_gp(X, y, **kwargs):
    feature_rms = 1.0 / np.sqrt(np.average(X**2, axis=0))
    feature_rms = np.nan_to_num(feature_rms, 1)
    y_mean = torch.sum(y) / torch.sum(X)
    mean, std = base_gp(
        X,
        y - (torch.sum(X, axis=1) * y_mean).reshape(y.shape),
        NormalizedDotProduct,
        {"diagonal_elements": feature_rms},
        **kwargs,
    )
    return mean + y_mean, std


def gp(X, y, **kwargs):
    return base_gp(
        X, y, DotProduct, {"sigma_0": 0, "sigma_0_bounds": "fixed"}, **kwargs
    )


def base_gp(
    X,
    y,
    kernel,
    kernel_kwargs,
    alpha: Optional[float] = 0.1,
    max_iteration: int = 20,
    stride: Optional[int] = 1,
):
    if len(y.shape) == 1:
        y = y.reshape([-1, 1])

    if stride is not None:
        X = X[::stride]
        y = y[::stride]

    not_fit = True
    iteration = 0
    mean = None
    std = None
    while not_fit:
        print(f"GP fitting iteration {iteration} {alpha}")
        try:
            _kernel = kernel(**kernel_kwargs)
            gpr = GaussianProcessRegressor(
                kernel=_kernel, random_state=0, alpha=alpha
            )  # noqa: E501
            gpr = gpr.fit(X, y)

            vec = torch.diag(torch.ones(X.shape[1]))
            mean, std = gpr.predict(vec, return_std=True)

            mean = torch.as_tensor(
                mean, dtype=torch.get_default_dtype()
            ).reshape(  # noqa: E501
                [-1]
            )  # noqa: E501
            # ignore all the off-diagonal terms
            std = torch.as_tensor(
                std, dtype=torch.get_default_dtype()
            ).reshape(  # noqa: E501
                [-1]
            )  # noqa: E501
            likelihood = gpr.log_marginal_likelihood()

            res = torch.sqrt(
                torch.square(torch.matmul(X, mean.reshape([-1, 1])) - y).mean()
            )

            print(
                f"GP fitting: alpha {alpha}:\n"
                f"            residue {res}\n"
                f"            mean {mean} std {std}\n"
                f"            log marginal likelihood {likelihood}"
            )
            not_fit = False

        except Exception as e:
            print(f"GP fitting failed for alpha={alpha} and {e.args}")
            if alpha == 0 or alpha is None:
                logging.info("try a non-zero alpha")
                not_fit = False
                raise ValueError(
                    f"Please set the {alpha} to non-zero value. \n"
                    "The dataset energy is rank deficient to be solved with GP"
                )
            else:
                alpha = alpha * 2
                iteration += 1
                logging.debug(f"           increase alpha to {alpha}")

            if iteration >= max_iteration or not_fit is False:
                raise ValueError(
                    "Please set the per species shift and scale "
                    "to zeros and ones. \nThe dataset energy is "
                    "to diverge to be solved with GP"
                )

    return mean, std


class NormalizedDotProduct(Kernel):
    r"""Dot-Product kernel.
    .. math::
        k(x_i, x_j) = x_i \cdot A \cdot x_j
    """

    def __init__(self, diagonal_elements):
        # TODO: check shape
        self.diagonal_elements = diagonal_elements
        self.A = np.diag(diagonal_elements)

    def __call__(self, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.
            Only supported when Y is None.
        Returns
        -------
        K : ndarray of shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : ndarray of shape (n_samples_X, n_samples_X, n_dims),\
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        X = np.atleast_2d(X)
        if Y is None:
            K = (X.dot(self.A)).dot(X.T)
        else:
            if eval_gradient:
                raise ValueError(
                    "Gradient can only be evaluated when Y is None."  # noqa: E501
                )
            K = (X.dot(self.A)).dot(Y.T)

        if eval_gradient:
            return K, np.empty((X.shape[0], X.shape[0], 0))
        else:
            return K

    def diag(self, X):
        """Returns the diagonal of the kernel k(X, X).
        The result of this method is identical to np.diag(self(X)); however,
        it can be evaluated more efficiently since only the diagonal is
        evaluated.
        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y).
        Returns
        -------
        K_diag : ndarray of shape (n_samples_X,)
            Diagonal of kernel k(X, X).
        """
        return np.einsum("ij,ij,jj->i", X, X, self.A)

    def __repr__(self):
        return ""

    def is_stationary(self):
        """Returns whether the kernel is stationary."""
        return False

    @property
    def hyperparameter_diagonal_elements(self):
        return Hyperparameter("diagonal_elements", "numeric", "fixed")



import argparse
import uuid
from datetime import datetime
from typing import List, Union

from ase import Atoms
from ase.io import read as ase_read
from loguru import logger

from mattersim.cli.applications.moldyn import moldyn
from mattersim.cli.applications.phonon import phonon
from mattersim.cli.applications.relax import relax
from mattersim.cli.applications.singlepoint import singlepoint
from mattersim.forcefield import MatterSimCalculator


def singlepoint_cli(args: argparse.Namespace) -> dict:
    """
    CLI wrapper for singlepoint function.

    Args:
        args (argparse.Namespace): Command line arguments.

    Returns:
        dict: Dictionary containing the predicted properties.

    """
    atoms_list = parse_atoms_list(
        args.structure_file, args.mattersim_model, args.device
    )
    singlepoint_args = {
        k: v
        for k, v in vars(args).items()
        if k not in ["structure_file", "mattersim_model", "device"]
    }
    return singlepoint(atoms_list, **singlepoint_args)


def relax_cli(args: argparse.Namespace) -> dict:
    """
    CLI wrapper for relax function.

    Args:
        args (argparse.Namespace): Command line arguments.

    Returns:
        dict: Dictionary containing the relaxed results
    """
    atoms_list = parse_atoms_list(
        args.structure_file, args.mattersim_model, args.device
    )
    relax_args = {
        k: v
        for k, v in vars(args).items()
        if k not in ["structure_file", "mattersim_model", "device"]
    }
    return relax(atoms_list, **relax_args)


def phonon_cli(args: argparse.Namespace) -> dict:
    """
    CLI wrapper for phonon function.

    Args:
        args (argparse.Namespace): Command line arguments.

    Returns:
        dict: Dictionary containing the phonon properties.
    """
    atoms_list = parse_atoms_list(
        args.structure_file, args.mattersim_model, args.device
    )
    phonon_args = {
        k: v
        for k, v in vars(args).items()
        if k not in ["structure_file", "mattersim_model", "device"]
    }
    return phonon(atoms_list, **phonon_args)


def moldyn_cli(args: argparse.Namespace) -> dict:
    """
    CLI wrapper for moldyn function.

    Args:
        args (argparse.Namespace): Command line arguments.

    Returns:
        dict: Dictionary containing the molecular dynamics properties.
    """
    atoms_list = parse_atoms_list(
        args.structure_file, args.mattersim_model, args.device
    )
    if len(atoms_list) > 1:
        logger.error("Molecular dynamics may take too long for multiple structures.")

    moldyn_args = {
        k: v
        for k, v in vars(args).items()
        if k not in ["structure_file", "mattersim_model", "device"]
    }
    return moldyn(atoms_list, **moldyn_args)


def add_common_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--structure-file",
        type=str,
        nargs="+",
        help="Path to the atoms structure file(s).",
    )
    parser.add_argument(
        "--mattersim-model",
        type=str,
        choices=["mattersim-v1.0.0-1m", "mattersim-v1.0.0-5m"],
        default="mattersim-v1.0.0-1m",
        help="MatterSim model to use.",
    )
    parser.add_argument(
        "--device",
        type=str,
        default="cpu",
        choices=["cpu", "cuda"],
        help="Device to use for prediction. Default is cpu.",
    )
    parser.add_argument(
        "--work-dir",
        type=str,
        default=datetime.now().strftime("%Y-%m-%d-%H-%M-%S") + "-" + str(uuid.uuid4()),
        help="Working directory for the calculations. "
        "Defaults to a UUID with timestamp when not set.",
    )
    parser.add_argument(
        "--save-csv",
        type=str,
        default="results.csv.gz",
        help="Save the results to a CSV file. "
        "Defaults to `results.csv.gz` when not set.",
    )


def add_relax_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--optimizer",
        type=str,
        default="FIRE",
        help="The optimizer to use. Default is FIRE.",
    )
    parser.add_argument(
        "--filter",
        type=str,
        default=None,
        help="The filter to use.",
    )
    parser.add_argument(
        "--constrain-symmetry",
        action="store_true",
        help="Constrain symmetry.",
    )
    parser.add_argument(
        "--fix-axis",
        type=bool,
        default=False,
        nargs="+",
        help="Fix the axis.",
    )
    parser.add_argument(
        "--pressure-in-GPa",
        type=float,
        default=None,
        help="Pressure in GPa to use for relaxation.",
    )
    parser.add_argument(
        "--fmax",
        type=float,
        default=0.01,
        help="Maximum force tolerance for relaxation.",
    )
    parser.add_argument(
        "--steps",
        type=int,
        default=500,
        help="Maximum number of steps for relaxation.",
    )


def add_phonon_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--find-prim",
        action="store_true",
        help="If find the primitive cell and use it to calculate phonon.",
    )
    parser.add_argument(
        "--amplitude",
        type=float,
        default=0.01,
        help="Magnitude of the finite difference to displace in "
        "force constant calculation, in Angstrom.",
    )
    parser.add_argument(
        "--supercell-matrix",
        type=int,
        nargs=3,
        default=None,
        help="Supercell matrix for construct supercell, must be a list of 3 integers.",
    )
    parser.add_argument(
        "--qpoints-mesh",
        type=int,
        nargs=3,
        default=None,
        help="Qpoint mesh for IBZ integral, must be a list of 3 integers.",
    )
    parser.add_argument(
        "--max-atoms",
        type=int,
        default=None,
        help="Maximum atoms number limitation for the supercell generation.",
    )
    parser.add_argument(
        "--enable-relax",
        action="store_true",
        help="Whether to relax the structure before predicting phonon properties.",
    )


def add_moldyn_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--temperature",
        type=float,
        default=300,
        help="Temperature in Kelvin.",
    )
    parser.add_argument(
        "--timestep",
        type=float,
        default=1,
        help="Timestep in femtoseconds.",
    )
    parser.add_argument(
        "--steps",
        type=int,
        default=1000,
        help="Number of steps for the molecular dynamics simulation.",
    )
    parser.add_argument(
        "--ensemble",
        type=str,
        choices=["nvt_berendsen", "nvt_nose_hoover"],
        default="nvt_nose_hoover",
        help="Simulation ensemble to use.",
    )
    parser.add_argument(
        "--logfile",
        type=str,
        default="md.log",
        help="Logfile to write the output to. Default is stdout.",
    )
    parser.add_argument(
        "--loginterval",
        type=int,
        default=10,
        help="Log interval for writing the output.",
    )
    parser.add_argument(
        "--trajectory",
        type=str,
        default="md.traj",
        help="Path to the trajectory file.",
    )
    parser.add_argument(
        "--taut",
        type=float,
        default=None,
        help="Taut in femtoseconds. It is used for the Nose-Hoover thermostat.",
    )


def parse_atoms_list(
    structure_file_list: Union[str, List[str]],
    mattersim_model: str,
    device: str = "cpu",
) -> List[Atoms]:
    if isinstance(structure_file_list, str):
        structure_file_list = [structure_file_list]

    calc = MatterSimCalculator(load_path=mattersim_model, device=device)
    atoms_list = []
    for structure_file in structure_file_list:
        atoms_list += ase_read(structure_file, index=":")
    for atoms in atoms_list:
        atoms.calc = calc
    return atoms_list


def main():
    argparser = argparse.ArgumentParser(description="CLI for MatterSim.")
    subparsers = argparser.add_subparsers(
        title="Subcommands",
        description="Valid subcommands",
        help="Available subcommands",
    )

    # Sub-command for single-point prediction
    singlepoint_parser = subparsers.add_parser(
        "singlepoint", help="Predict single point properties for a list of atoms."
    )
    add_common_args(singlepoint_parser)
    singlepoint_parser.set_defaults(func=singlepoint_cli)

    # Sub-command for relax
    relax_parser = subparsers.add_parser(
        "relax", help="Relax a list of atoms structures."
    )
    add_common_args(relax_parser)
    add_relax_args(relax_parser)
    relax_parser.set_defaults(func=relax_cli)

    # Sub-command for phonon
    phonon_parser = subparsers.add_parser(
        "phonon",
        help="Predict phonon properties for a list of structures.",
    )
    add_common_args(phonon_parser)
    add_relax_args(phonon_parser)
    add_phonon_args(phonon_parser)
    phonon_parser.set_defaults(func=phonon_cli)

    # Sub-command for molecular dynamics
    moldyn_parser = subparsers.add_parser(
        "moldyn",
        help="Perform molecular dynamics simulation for a list of structures.",
    )
    add_common_args(moldyn_parser)
    add_moldyn_args(moldyn_parser)
    moldyn_parser.set_defaults(func=moldyn_cli)

    # Parse arguments
    args = argparser.parse_args()
    print(args)

    # Call the function associated with the sub-command
    if hasattr(args, "func"):
        args.func(args)
    else:
        argparser.print_help()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from typing import Union

from ase import Atoms, units
from ase.io import Trajectory
from ase.md.npt import NPT
from ase.md.nvtberendsen import NVTBerendsen
from ase.md.velocitydistribution import (  # noqa: E501
    MaxwellBoltzmannDistribution,
    Stationary,
)


class MolecularDynamics:
    """
    This class is used for Molecular Dynamics.
    """

    SUPPORTED_ENSEMBLE = ["NVT_BERENDSEN", "NVT_NOSE_HOOVER"]

    def __init__(
        self,
        atoms: Atoms,
        ensemble: str = "nvt_nose_hoover",
        temperature: float = 300,
        timestep: float = 1.0,
        taut: Union[float, None] = None,
        trajectory: Union[str, Trajectory, None] = None,
        logfile: Union[str, None] = "-",
        loginterval: int = 10,
        append_trajectory: bool = False,
    ):
        """
        Args:
            atoms (Union[Atoms, Structure]): ASE atoms object contains
                structure information and calculator.
            ensemble (str, optional): Simulation ensemble chosen. Defaults
                to nvt_nose_hoover'
            temperature (float, optional): Simulation temperature, in Kelvin.
                Defaults to 300 K.
            timestep (float, optional): The simulation time step, in fs. Defaults
                to 1 fs.
            taut (float, optional): Characteristic timescale of the thermostat,
                in fs. If is None, automatically set it to 1000 * timestep.
            trajectory (Union[str, Trajectory], optional): Attach trajectory
                object. If trajectory is a string a Trajectory will be constructed.
                Defaults to None, which means for no trajectory.
            logfile (str, optional): If logfile is a string, a file with that name
                will be opened. Defaults to '-', which means output to stdout.
            loginterval (int, optional): Only write a log line for every loginterval
                time steps. Defaults to 10.
            append_trajectory (bool, optional): If False the trajectory file to be
                overwritten each time the dynamics is restarted from scratch. If True,
                the new structures are appended to the trajectory file instead.

        """
        assert atoms.calc is not None, (
            "Molecular Dynamics simulation only accept "
            "ase atoms with an attached calculator"
        )
        if ensemble.upper() not in self.SUPPORTED_ENSEMBLE:
            raise NotImplementedError(  # noqa: E501
                f"Ensemble {ensemble} is not yet supported."
            )

        self.atoms = atoms

        self.ensemble = ensemble.upper()
        self._temperature = temperature
        self.timestep = timestep

        if taut is None:
            taut = 1000 * timestep * units.fs
        self.taut = taut

        self._trajectory = trajectory
        self.logfile = logfile
        self.loginterval = loginterval
        self.append_trajectory = append_trajectory

        self._initialize_dynamics()

    def _initialize_dynamics(self):
        """
        Initialize the Dynamic ensemble class.
        """
        MaxwellBoltzmannDistribution(
            self.atoms, temperature_K=self._temperature, force_temp=True
        )
        Stationary(self.atoms)

        if self.ensemble == "NVT_BERENDSEN":  # noqa: E501
            self.dyn = NVTBerendsen(
                self.atoms,
                timestep=self.timestep * units.fs,
                temperature_K=self._temperature,
                taut=self.taut,
                trajectory=self._trajectory,
                logfile=self.logfile,
                loginterval=self.loginterval,
                append_trajectory=self.append_trajectory,
            )
        elif self.ensemble == "NVT_NOSE_HOOVER":
            self.dyn = NPT(
                self.atoms,
                timestep=self.timestep * units.fs,
                temperature_K=self._temperature,
                ttime=self.taut,
                pfactor=None,
                trajectory=self._trajectory,
                logfile=self.logfile,
                loginterval=self.loginterval,
                append_trajectory=self.append_trajectory,
            )
        else:
            raise NotImplementedError(  # noqa: E501
                f"Ensemble {self.ensemble} is not yet supported."
            )

    def run(self, n_steps: int = 1):
        """
        Run Molecular Dynamic simulation.

        Args:
            n_steps (int, optional): Number of steps to simulations. Defaults to 1.
        """
        self.dyn.run(n_steps)

    @property
    def temperature(self):
        return self._temperature

    @temperature.setter
    def temperature(self, temperature: float):
        self._temperature = temperature
        self._initialize_dynamics()

    @property
    def trajectory(self):
        return self._trajectory

    @trajectory.setter
    def trajectory(self, trajectory: Union[str, Trajectory, None]):
        self._trajectory = trajectory
        self._initialize_dynamics()



"""
Tests for __version__.py (GitHub issue #140):
Mattersim failed to import with setuptools>=82 due to removed pkg_resources.
"""

import importlib
import sys
from unittest import mock


def _reimport_version_module():
    """Force a fresh import of mattersim.__version__."""
    for mod_name in list(sys.modules):
        if mod_name.startswith("mattersim"):
            del sys.modules[mod_name]
    return importlib.import_module("mattersim.__version__")


def test_version_available():
    """mattersim.__version__ is a non-empty string."""
    import mattersim

    assert isinstance(mattersim.__version__, str)
    assert mattersim.__version__


def test_import_succeeds_without_pkg_resources():
    """Import must work when pkg_resources is absent (setuptools>=82, issue #140)."""
    for mod_name in list(sys.modules):
        if mod_name.startswith("mattersim"):
            del sys.modules[mod_name]

    with mock.patch.dict(sys.modules, {"pkg_resources": None}):
        mod = importlib.import_module("mattersim.__version__")
        assert isinstance(mod.__version__, str)
        assert mod.__version__



# -*- coding: utf-8 -*-
import unittest

from ase import Atoms
from ase.calculators.emt import EMT
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from mattersim.applications.relax import Relaxer


class RelaxerTestCase(unittest.TestCase):
    def setUp(self):
        # Create an example structure with displaced atoms for testing
        a = 1.786854996  # Angstroms
        positions = [
            (0, 0, 0),
            (a / 4, a / 4, a / 4),
            (a / 2, a / 2, 0),
            (a / 2, 0, a / 2),
            (0, a / 2, a / 2),
            (a / 4, 3 * a / 4, 3 * a / 4.01),  # displaced
            (3 * a / 4, a / 4.01, 3 * a / 4),  # displaced
            (3 * a / 4, 3 * a / 4, a / 4),
        ]
        cell = [(a, 0, 0), (0, a, 0), (0, 0, a)]
        self.atoms_displaced = Atoms(
            "C8", positions=positions, cell=cell, pbc=True  # noqa: E501
        )

        # Create an example structure with expanded cell for testing
        a = 1.786854996 * 1.2
        positions = [
            (0, 0, 0),
            (a / 4, a / 4, a / 4),
            (a / 2, a / 2, 0),
            (a / 2, 0, a / 2),
            (0, a / 2, a / 2),
            (a / 4, 3 * a / 4, 3 * a / 4),
            (3 * a / 4, a / 4, 3 * a / 4),
            (3 * a / 4, 3 * a / 4, a / 4),
        ]
        cell = [(a, 0, 0), (0, a, 0), (0, 0, a)]
        self.atoms_expanded = Atoms(
            "C8", positions=positions, cell=cell, pbc=True  # noqa: E501
        )

        self.calculator = EMT()

    def test_default_relaxer(self):
        relaxer = Relaxer()
        atoms_displaced = self.atoms_displaced.copy()
        atoms_displaced.set_calculator(self.calculator)
        converged, relaxed_atoms = relaxer.relax(
            atoms_displaced, fmax=0.1, steps=500
        )  # noqa: E501
        self.assertTrue(converged)
        self.assertIsInstance(relaxed_atoms, Atoms)

    def test_relax_structures(self):
        atoms_list = [
            self.atoms_displaced.copy(),
            self.atoms_displaced.copy(),
            self.atoms_displaced.copy(),
        ]
        for atoms in atoms_list:
            atoms.set_calculator(self.calculator)

        converged_list, relaxed_atoms_list = Relaxer.relax_structures(
            atoms_list, fmax=0.1, steps=500
        )
        self.assertIsInstance(converged_list, list)
        for converged in converged_list:
            self.assertTrue(converged)

    def test_relax_structures_under_pressure(self):
        atoms_displaced = self.atoms_displaced.copy()
        atoms_displaced.set_calculator(self.calculator)
        init_volume = atoms_displaced.get_volume()
        print(f"Initial volume: {init_volume}")

        # First, relax under 0 pressure
        converged, relaxed_atoms = Relaxer.relax_structures(
            atoms_displaced,
            steps=500,
            fmax=0.1,
            filter="FrechetCellFilter",
            pressure_in_GPa=0.0,
        )
        intermediate_volume = relaxed_atoms.get_volume()
        print(f"Intermediate volume: {intermediate_volume}")
        self.assertTrue(converged)

        # Second, relax under 100 GPa
        converged, relaxed_atoms = Relaxer.relax_structures(
            relaxed_atoms,
            steps=500,
            fmax=0.1,
            filter="FrechetCellFilter",
            pressure_in_GPa=100.0,
        )
        final_volume = relaxed_atoms.get_volume()
        print(f"Final volume: {final_volume}")
        self.assertTrue(converged)
        self.assertLess(final_volume, intermediate_volume)
        print(f"Final cell: {relaxed_atoms.cell}")

    def test_relax_with_filter_and_constrained_symmetry(self):
        atoms_expanded = self.atoms_expanded.copy()
        atoms_expanded.set_calculator(self.calculator)
        init_volume = atoms_expanded.get_volume()
        print(f"Initial volume: {init_volume}")

        init_analyzer = SpacegroupAnalyzer(
            AseAtomsAdaptor.get_structure(self.atoms_expanded)
        )
        init_spacegroup = init_analyzer.get_space_group_number()

        # First, relax under 0 pressure
        converged, relaxed_atoms = Relaxer.relax_structures(
            atoms_expanded,
            steps=500,
            fmax=0.1,
            filter="FrechetCellFilter",
            pressure_in_GPa=0.0,
            constrain_symmetry=True,
        )
        intermediate_volume = relaxed_atoms.get_volume()
        print(f"Intermediate volume: {intermediate_volume}")
        self.assertTrue(converged)

        # Second, relax under 100 GPa
        converged, relaxed_atoms = Relaxer.relax_structures(
            relaxed_atoms,
            steps=500,
            fmax=0.1,
            filter="FrechetCellFilter",
            pressure_in_GPa=100.0,
            constrain_symmetry=True,
        )
        final_volume = relaxed_atoms.get_volume()
        print(f"Final volume: {final_volume}")
        self.assertTrue(converged)
        self.assertLess(final_volume, intermediate_volume)

        final_analyzer = SpacegroupAnalyzer(
            AseAtomsAdaptor.get_structure(relaxed_atoms)
        )
        final_spacegroup = final_analyzer.get_space_group_number()
        self.assertEqual(init_spacegroup, final_spacegroup)
        print(f"Final cell: {relaxed_atoms.cell}")
        cell_a = relaxed_atoms.cell[0, 0]
        cell_b = relaxed_atoms.cell[1, 1]
        cell_c = relaxed_atoms.cell[2, 2]
        self.assertAlmostEqual(cell_a, cell_b)
        self.assertAlmostEqual(cell_a, cell_c)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import numpy as np
from ase import Atoms
from ase.spacegroup.symmetrize import check_symmetry


def auto_grid_detection(
    atom: Atoms,
    max_atoms: int,
    ratio_tolerance: float = 1.1,
    is_santity_check: bool = True,
    is_verbose: bool = True,
):
    """
    This function automates the detection of grid for a given atomic structure
    and max_atoms. If lattice vectors length in three direction is same or the
    difference is smaller than 0.1, the supercell vector element will has the
    same value in three direction, the value is (max_atoms/atoms)^(1/3). Else
    the supercell vector element will be set proportionally to make the three
    supercell lattice vector length as same as possible.

    Args:
        atom (Atoms): ASE atoms object to provide lattice informations.
        max_atoms: (int): Maximum atom number limitation for supercell.
        ratio_tolerance (float, optional): The tolerance for the ratio of the
            lengths of the lattice vectors. Defaults to 1.1.
        is_santity_check (bool, optional): If True, performs a sanity check to
            ensure symmetry is preserved after replications. Defaults to True.
        is_verbose (bool, optional): If True, prints detailed information about
            the atomic structure and the replication process. Defaults to True.
    """
    # Get the cell length
    lattice_vector_lengths = atom.cell.cellpar()[:3]

    # Base case, absolute the same length
    if (
        lattice_vector_lengths[0]
        == lattice_vector_lengths[1]
        == lattice_vector_lengths[2]
    ):
        number_of_replicas = int(np.round(max_atoms / (len(atom))) ** (1 / 3))
        number_of_replicas = max(number_of_replicas, 1)
        max_replication = (
            number_of_replicas,
            number_of_replicas,
            number_of_replicas,
        )

    # Case 1: Non-symmetry lengths within tolerance
    else:
        lattice_vector_lengths_argsort_indices = np.argsort(lattice_vector_lengths)[
            ::-1
        ]
        sorted_lattice_vector_lengths = lattice_vector_lengths[
            lattice_vector_lengths_argsort_indices
        ]
        ratios = [
            sorted_lattice_vector_lengths[0] / sorted_lattice_vector_lengths[1],
            sorted_lattice_vector_lengths[0] / sorted_lattice_vector_lengths[2],
        ]

        # Variation in cell length with in the tolerance we still consider it
        # as an N-N-N replications
        if ratios[0] <= ratio_tolerance and ratios[1] <= ratio_tolerance:
            number_of_replicas = int(np.round(max_atoms / (len(atom))) ** (1 / 3))
            number_of_replicas = max(number_of_replicas, 1)
            max_replication = (
                number_of_replicas,
                number_of_replicas,
                number_of_replicas,
            )

        # Case 2: Non-symmetry lengths beyond tolerance
        else:
            # Compute the replica along the most asymmetric direction
            asymmetric_replica = int(
                (max_atoms / len(atom) / np.prod(ratios)) ** (1 / 3)
            )
            asymmetric_replica = max(asymmetric_replica, 1)

            # Recover replica on the other two direction based on ratios
            replica_r0 = max(int(np.round(asymmetric_replica * ratios[0])), 1)
            replica_r1 = max(int(np.round(asymmetric_replica * ratios[1])), 1)
            indices_to_recover_lattice_vector_order = np.argsort(
                lattice_vector_lengths_argsort_indices
            )
            max_replication_arr = np.array(
                [asymmetric_replica, replica_r0, replica_r1]
            )[indices_to_recover_lattice_vector_order]
            max_replication = tuple(max_replication_arr)

    # Broad cast unit cell infomraiton
    if is_verbose:
        print("System:", atom)
        print("Number of atoms in the unit cell: ", len(atom))
        print("Lattice vector and angles: ", atom.cell.cellpar())
        print(
            "Space group: ",
            check_symmetry(atom, 1e-3, verbose=False)["international"],
        )

    if is_santity_check:
        symmetry_of_unit_cell = check_symmetry(atom, 1e-3, verbose=False)[
            "international"
        ]
        symmetry_of_replicated_supercell = check_symmetry(
            atom.copy().repeat(max_replication), 1e-3, verbose=False
        )["international"]
        if symmetry_of_unit_cell == symmetry_of_replicated_supercell:
            print(
                "symmetry is preserved after replications, safely return "
                "replication combination !\n"
            )
            return max_replication
        else:
            print(
                "Symmetry is lose after replications. No possible replications"
                " can be found !\n"
            )
            return (1, 1, 1)

    # Check if max_replication is still the initial value
    if max_replication == (1, 1, 1):
        print("No possible replications. Returning unit cell.")
        return (1, 1, 1)
    else:
        return max_replication


def get_supercell_parameters(
    atom: Atoms,
    supercell_matrix: np.ndarray = None,
    qpoints_mesh: np.ndarray = None,
    max_atoms: int = None,
):
    """
    Based on symmetry to get supercell setting parameters.
    First, setting the maximum atoms number limitation for supercell. If max_atoms
    is None, will automatic setting it, else use user assigned. If the lattice
    parameters in three direction is same or approximately same, the max_atoms will
    be set small, e.g. 216 or 300; else some direction needed expand larger than
    others, so the max_atoms also need more, e.g. 450. Then, based on max_atoms,
    call auto_grid_dection function to obtain supercell matrix diagonal elements.
    Finally setting the k_point_mesh used to integrate Brillouin Zone. Cause kpoints
    in inverse space, smaller real space means the inverse space is larger, will
    need more kpoints.

    Args:
        atom (Atoms): ASE atoms object to provide lattice information.
        supercell_matrix (nd.ndarray, optional): Supercell matrix for construct
            supercell, prior than max_atoms.
        qpoints_mesh (np.ndarray, optional): Qpoints mesh for IBZ integral, prio
            over than max_atoms.
        max_atoms (int, optional): If not None, will use user setting maximum
            atoms number limitation for generate supercell, else automatic set.
            Defaults to None.
    """
    if supercell_matrix is not None:
        nrep_second = np.diag(supercell_matrix)
        if nrep_second[0] == nrep_second[1] == nrep_second[2]:
            k_point_mesh = 6 * np.array(nrep_second)
        else:
            k_point_mesh = 3 * np.array(nrep_second)

        if qpoints_mesh is not None:
            k_point_mesh = qpoints_mesh

        return supercell_matrix, k_point_mesh

    lattice_vector_lengths = atom.cell.cellpar()[:3]
    lattice_vector_lengths_argsort_indices = np.argsort(lattice_vector_lengths)[::-1]
    sorted_lattice_vector_lengths = lattice_vector_lengths[
        lattice_vector_lengths_argsort_indices
    ]
    ratios = [
        sorted_lattice_vector_lengths[0] / sorted_lattice_vector_lengths[1],
        sorted_lattice_vector_lengths[0] / sorted_lattice_vector_lengths[2],
    ]
    if max_atoms:
        pass
    elif (
        check_symmetry(atom, 1e-3, verbose=False)["international"] == "Fd-3m"
        or check_symmetry(atom, 1e-3, verbose=False)["international"] == "Fm-3m"
        or check_symmetry(atom, 1e-3, verbose=False)["international"] == "F-43m"
    ):
        max_atoms = 216
    elif check_symmetry(atom, 1e-3, verbose=False)["international"] == "P6_3mc":
        max_atoms = 450
    elif ratios[0] <= 1.1 and ratios[1] <= 1.1:
        max_atoms = 300
    else:
        max_atoms = 300

    nrep_second = auto_grid_detection(atom, max_atoms, is_verbose=False)

    if nrep_second[0] == nrep_second[1] == nrep_second[2]:
        k_point_mesh = 6 * np.array(nrep_second)
    else:
        k_point_mesh = 3 * np.array(nrep_second)

    return nrep_second, k_point_mesh
